# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import GObject, GLib

import logging
import os
import shutil
from threading import Thread
from typing import Optional

import pypandoc

from iotas.attachment_helpers import (
    AttachmentsCopyOutcome,
    copy_note_attachments,
    get_attachment_disk_states,
    get_attachments_dir,
)
from iotas.export_utils import (
    build_default_filename,
    get_container_export_dir,
    sanitise_title_for_filename,
)
from iotas.html_generator import HtmlGenerator
from iotas.markdown_helpers import (
    get_image_attachments_from_note_content,
    get_note_export_content,
    parse_to_tokens,
)
from iotas.note import Note
from iotas.note_exporter import NoteExporter
from iotas.pdf_exporter import PdfExporter


class DefaultNoteExporter(NoteExporter):
    """Note exporter.

    :param PdfExporter pdf_exporter: PDF exporter
    :param HtmlGenerator html_generator: HTML generator
    :param str app_data_path: User data path
    """

    __gsignals__ = {
        # str: single export out path
        "finished": (GObject.SignalFlags.RUN_FIRST, None, (str,)),
        # str: reason
        "failed": (GObject.SignalFlags.RUN_FIRST, None, (str,)),
    }

    def __init__(
        self, pdf_exporter: PdfExporter, html_generator: HtmlGenerator, app_data_path: str
    ) -> None:
        super().__init__()
        self.__pdf_exporter = pdf_exporter
        self.__pdf_exporter.set_callbacks(
            self.__on_pdf_export_finished, self.__on_pdf_export_failed
        )
        self.__html_generator = html_generator
        self.__app_data_path = app_data_path
        self.__allow_missing_images = False

    def export(
        self,
        note: Note,
        out_format: str,
        file_extension: str,
        supporting_tex: bool,
        allow_missing_images: bool = False,
        user_location: Optional[str] = None,
    ) -> None:
        """Export note.

        :param Note note: Note to render
        :param str out_format: Export format
        :param str file_extension: File extension
        :param bool supporting_tex: TeX support
        :param bool allow_missing_images: Whether missing images are treated as a failure
        :param str user_location: User chosen export location, optional
        """
        self.__note = note
        self.__out_format = out_format
        self.__supporting_tex = supporting_tex
        self.__allow_missing_images = allow_missing_images

        if user_location:
            self.__location = user_location
        else:
            # Running with limited permissions in container, export to exports dir inside container
            # with automatic filename
            export_dir = get_container_export_dir()
            filename = build_default_filename(note, out_format, file_extension, add_timestamp=True)
            self.__location = os.path.join(export_dir, filename)
            if not os.path.exists(export_dir):
                try:
                    os.mkdir(export_dir)
                except OSError as e:
                    logging.warning(f"Failed to export {out_format} to {self.__location}: {e}")
                    self.emit("failed", e)
                    return

        if not self.__check_for_missing_attachments():
            return

        if out_format == "pdf":
            self.__export_pdf()
        elif out_format == "md":
            self.__export_md()
        elif out_format == "html":
            self.__export_html()
        else:
            logging.info(f"Asking pandoc to export to {out_format}")
            self.__export_pandoc(out_format)

    def __on_pdf_export_finished(self) -> None:
        GLib.idle_add(self.emit, "finished", self.__location)

    def __on_pdf_export_failed(self, error: str) -> None:
        GLib.idle_add(self.emit, "failed", error)

    def __export_pandoc(self, out_format: str) -> None:
        content = get_note_export_content(self.__note, prefix_note_id=True)
        working_dir = os.path.abspath(os.path.join(get_attachments_dir(), os.pardir))

        def thread_do():
            try:
                pypandoc.convert_text(
                    content,
                    out_format,
                    format="gfm+attributes+implicit_figures",
                    outputfile=self.__location,
                    cworkdir=working_dir,
                )
            except (RuntimeError, OSError) as e:
                logging.warning(f"Failed to export {out_format} to {self.__location}: {e}")
                GLib.idle_add(self.emit, "failed", str(e))
            else:
                logging.info(f"Exported {self.__out_format} to {self.__location}")
                GLib.idle_add(self.emit, "finished", self.__location)

        thread = Thread(target=thread_do)
        thread.daemon = True
        thread.start()

    def __export_pdf(self) -> None:
        self.__pdf_exporter.export(self.__note, self.__location)

    def __export_md(self) -> None:

        def thread_do() -> None:
            if len(get_image_attachments_from_note_content(self.__note)) > 0:
                file_location = os.path.join(
                    self.__location, f"{sanitise_title_for_filename(self.__note.title)}.md"
                )
                if not self.__create_export_directory():
                    return
            else:
                file_location = self.__location

            export_content = get_note_export_content(self.__note, prefix_note_id=False)
            try:
                with open(file_location, "w") as f:
                    f.write(export_content)
            except OSError as e:
                GLib.idle_add(self.emit, "failed", e)
                logging.warning(f"Failed to export MD: {e}")
                return

            if not self.__copy_attachments():
                return

            logging.info(f"Exported {self.__out_format} to {file_location}")
            GLib.idle_add(self.emit, "finished", self.__location)

        thread = Thread(target=thread_do)
        thread.daemon = True
        thread.start()

    def __export_html(self) -> None:

        def thread_do() -> None:

            parser, tokens = parse_to_tokens(
                self.__note, exporting=True, tex_support=self.__supporting_tex
            )
            html = self.__html_generator.generate(
                self.__note,
                tokens,
                parser.renderer.render,
                parser.options,
                searching=False,
                export_format="html",
            )

            if not self.__create_export_directory():
                return

            index_filename = os.path.join(self.__location, "index.html")

            try:
                with open(index_filename, "w") as f:
                    f.write(html)
            except OSError as e:
                GLib.idle_add(self.emit, "failed", e)
                logging.warning(f"Failed to export HTML: {e}")
                return

            if not self.__copy_attachments():
                return

            css_dir = os.path.join(self.__location, "css")

            try:
                os.mkdir(css_dir)
            except OSError as e:
                GLib.idle_add(self.emit, "failed", e)
                logging.warning(f"Failed to export HTML: {e}")
                return

            dest_file = os.path.join(css_dir, os.path.basename(HtmlGenerator.RESOURCE_CSS_PATH))

            try:
                shutil.copyfile(
                    f"{self.__app_data_path}/{HtmlGenerator.RESOURCE_CSS_PATH}", dest_file
                )
            except OSError as e:
                GLib.idle_add(self.emit, "failed", e)
                logging.warning(f"Failed to export HTML: {e}")
                return

            if self.__supporting_tex:
                dest_file = os.path.join(
                    css_dir, os.path.basename(HtmlGenerator.RESOURCE_KATEX_CSS_PATH)
                )

                try:
                    shutil.copyfile(
                        f"{self.__app_data_path}/{HtmlGenerator.RESOURCE_KATEX_CSS_PATH}", dest_file
                    )
                except OSError as e:
                    GLib.idle_add(self.emit, "failed", e)
                    logging.warning(f"Failed to export HTML: {e}")
                    return

                js_dir = os.path.join(self.__location, "js")

                try:
                    os.mkdir(js_dir)
                except OSError as e:
                    GLib.idle_add(self.emit, "failed", e)
                    logging.warning(f"Failed to export HTML: {e}")
                    return

                dest_file = os.path.join(
                    js_dir, os.path.basename(HtmlGenerator.RESOURCE_KATEX_JS_PATH)
                )

                try:
                    shutil.copyfile(
                        f"{self.__app_data_path}/{HtmlGenerator.RESOURCE_KATEX_JS_PATH}", dest_file
                    )
                except OSError as e:
                    GLib.idle_add(self.emit, "failed", e)
                    logging.warning(f"Failed to export HTML: {e}")
                    return

            logging.info(f"Exported {self.__out_format} to {self.__location}")
            GLib.idle_add(self.emit, "finished", self.__location)

        thread = Thread(target=thread_do)
        thread.daemon = True
        thread.start()

    def __create_export_directory(self) -> bool:
        if os.path.exists(self.__location):
            logging.debug(f"Removing existing {self.__location}")
            try:
                shutil.rmtree(self.__location)
            except OSError as e:
                self.emit("failed", e)
                logging.warning(f"Failed to replace dir for {self.__out_format} export: {e}")
                return False

        try:
            os.mkdir(self.__location)
        except OSError as e:
            self.emit("failed", e)
            logging.warning(f"Failed to make dir for {self.__out_format} export: {e}")
            return False

        return True

    def __check_for_missing_attachments(self) -> bool:
        if len(get_image_attachments_from_note_content(self.__note)) > 0:
            logging.debug("Checking attachments are on disk")
            parser, tokens = parse_to_tokens(
                self.__note, exporting=True, tex_support=self.__supporting_tex
            )
            states = get_attachment_disk_states(self.__note, tokens)
            success = len(states.missing) == 0
            if not success and not self.__allow_missing_images:
                self.emit("failed", "Failed to export some attachments")
                return False

        return True

    def __copy_attachments(self) -> bool:
        if len(get_image_attachments_from_note_content(self.__note)) > 0:
            attachments_dir = os.path.join(self.__location, "attachments")
            logging.debug("Copying attachments")
            result = copy_note_attachments(self.__note, attachments_dir, prefix_note_id=False)
            success = result.outcome in (
                AttachmentsCopyOutcome.NONE,
                AttachmentsCopyOutcome.SUCCESS,
            )
            if not success and not self.__allow_missing_images:
                self.emit("failed", "Failed to export some attachments")
                return False

        return True



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from abc import ABC, abstractmethod

from typing import Callable

from iotas.note import Note


class PdfExporter(ABC):
    """PDF export interface."""

    @abstractmethod
    def set_callbacks(self, finished_callback: Callable, error_callback: Callable) -> None:
        """Set functions to be called upon export result.

        :param Callable finished_callback: Finished callback
        :param Callable error_callback: Error callback
        """
        raise NotImplementedError()

    @abstractmethod
    def export(self, note: Note, location: str) -> None:
        """Export PDF of note.

        :param Note note: Note to export
        :param str location: Destination location
        """
        raise NotImplementedError()



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gettext import gettext as _
import gi

gi.require_version("Adw", "1")
gi.require_version("GtkSource", "5")
from gi.repository import Adw, Gdk, Gio, GLib, GObject, Gtk, GtkSource

import logging
from typing import Callable, Optional

from iotas.attachment_helpers import flush_all_attachments_on_disk
from iotas.backup_manager import BackupManager
import iotas.const as const
from iotas.config_manager import ConfigManager
from iotas.database import Database, DbCursor
from iotas.migration_assistant import MigrationAssistant
from iotas.note_database import NoteDatabase
from iotas.preferences_dialog import PreferencesDialog
from iotas.nextcloud_sync_worker import NextcloudSyncWorker
from iotas.sync_manager import SyncManager
from iotas.widgets import load_widgets
from iotas.window import Window


class Application(Adw.Application):
    development_mode = const.IS_DEVEL
    application_id = const.APP_ID

    def __init__(self, *args) -> None:
        super().__init__(
            *args,
            application_id=self.application_id,
            flags=Gio.ApplicationFlags.HANDLES_OPEN,
            resource_base_path="/org/gnome/World/Iotas",
            register_session=True,
        )

        ConfigManager.set_app_id(const.APP_ID)
        self.__config_manager = ConfigManager.get_default()

        self.__add_cli_options()

        self.__index_category_label_css_provider: Gtk.CssProvider = None
        style_manager = Adw.StyleManager.get_default()
        style_manager.connect("notify::dark", lambda _o, _v: self.__apply_css())
        style_manager.connect("notify::high-contrast", lambda _o, _v: self.__apply_css())

        self.cursors: dict[str, DbCursor] = {}
        self.__actions: dict[str, Gio.SimpleAction] = {}
        self.__window: Optional[Adw.ApplicationWindow] = None
        self.__preferences_dialog: Optional[PreferencesDialog] = None
        self.__previous_version = ""

        self.connect("startup", lambda _o: self.__on_startup())
        self.connect("activate", lambda _o: self.__on_activate())
        self.connect("handle-local-options", self.__on_handle_local_options)

        self.__config_manager.connect_changed(
            ConfigManager.INDEX_CATEGORY_STYLE,
            self.__apply_index_category_label_css,
        )

    def disconnect_nextcloud(self) -> None:
        """Disconnect from Nextcloud instance and exit."""
        if not self.__sync_manager.authenticated:
            return
        assert self.__window  # mypy

        self.__sync_manager.sign_out()
        self.db.delete_all_clean_synced_notes()
        flush_all_attachments_on_disk()
        self.__config_manager.nextcloud_username = ""
        self.__window.close()

    def do_open(self, files: list[Gio.File], n_files: int, _hint: str) -> None:
        """Handle opening "files". In this case URI for the temporary handler.

        :param list[Gio.File] files: Files
        :param int n_files: Number of files
        :param str _hint: Hint
        """

        # Note: Iotas' addition of this URI handler is a stopgap measure until the freedesktop
        # intent system is fleshed out. The handling provided here will be unceremoniously removed
        # when that system is available and may be removed or modified without notice beforehand;
        # this is not a stable public interface.

        if n_files != 1:
            self.quit()
            return
        file = files[0]
        uri = GLib.Variant("s", file.get_uri())
        self.activate_action("create-note-from-uri", uri)

    def reset_sync_marker(self) -> None:
        """Resets the timestamp for transfer with Nextcloud Notes."""
        self.__sync_manager.reset_marker()

    def reset_database(self) -> None:
        """Destroy the database and quit."""
        self.db_base.trash()
        flush_all_attachments_on_disk()
        self.__config_manager.nextcloud_prune_threshold = 0
        self.quit()

    def apply_style(self) -> None:
        """Apply style preference."""
        manager = Adw.StyleManager.get_default()
        style = self.__config_manager.style
        if style == "dark":
            manager.props.color_scheme = Adw.ColorScheme.FORCE_DARK
        elif style == "light":
            manager.props.color_scheme = Adw.ColorScheme.FORCE_LIGHT
        else:
            manager.props.color_scheme = Adw.ColorScheme.DEFAULT

    @GObject.Property(type=bool, default=False)
    def debug_session(self) -> bool:
        return self.__debug_session

    @GObject.Property(type=str)
    def previous_version(self) -> str:
        return self.__previous_version

    @previous_version.setter
    def set_previous_version(self, value: str) -> None:
        self.__previous_version = value

    def __on_startup(self) -> None:
        """Handle startup."""
        Gtk.Application.do_startup(self)
        Adw.init()
        GtkSource.init()

        GtkSource.LanguageManager.get_default().prepend_search_path(
            "resource:///org/gnome/World/Iotas/gtksourceview-5/language-specs/"
        )
        GtkSource.StyleSchemeManager.get_default().prepend_search_path(
            "resource:///org/gnome/World/Iotas/gtksourceview-5/styles/"
        )

        devel = " (development build)" if const.IS_DEVEL else ""
        logging.info(f"Iotas v{const.VERSION}{devel}")

        self.db_base = Database()
        self.db = NoteDatabase(self.db_base)
        self.__sync_manager = SyncManager(self.db)
        self.__backup_manager = BackupManager(self.db, const.VERSION)

        self.previous_version = self.__config_manager.last_launched_version
        migration_assistant = MigrationAssistant(self.db)
        migration_assistant.migrate()

        load_widgets()
        self.__setup_actions()

    def __on_handle_local_options(self, _obj: GObject.Object, options: GLib.VariantDict) -> int:
        """Handle options, setup logging."""
        options = options.end().unpack()

        # Print paths before logging is setup and app initialised, for clean output
        if "display-backup-path" in options:
            print(BackupManager.get_default_primary_path())
            self.quit()
            return 0
        elif "display-ca-file-path" in options:
            print(NextcloudSyncWorker.CA_CHAIN_FILE)
            self.quit()
            return 0

        self.__debug_session = "debug-session" in options

        loglevel = logging.INFO
        if self.development_mode or self.__debug_session:
            loglevel = logging.DEBUG

            for module in ("urllib3", "gtkspellcheck", "markdown_it", "pypandoc"):
                logging.getLogger(module).setLevel(logging.INFO)

        logging.basicConfig(
            format="%(asctime)s | %(module)s | %(levelname)s | %(message)s",
            datefmt="%H:%M:%S",
            level=loglevel,
        )

        if "create-backup" in options:
            self.register()
            if self.get_property("is-remote"):
                self.activate_action("create-backup")
                logging.info("Requesting backup from already running Iotas")
                logging.info(
                    "Note: backup outcome will be logged from the other process; the return code "
                    "here doesn't reflect the backup result"
                )
                success = True
            else:
                success = self.__backup_manager.create_backup(
                    self.__config_manager.nextcloud_sync_configured,
                    self.__config_manager.backup_note_extension,
                )
            self.quit()
            return int(not success)
        elif "restore-backup" in options:
            self.register()
            if self.get_property("is-remote"):
                logging.error(
                    "Please quit the running instance of Iotas before restoring the backup"
                )
                self.quit()
                success = False
            else:
                success = self.__backup_manager.restore_backup(
                    self.__config_manager.nextcloud_sync_configured
                )
            self.quit()
            return int(not success)
        elif "quit-running" in options:
            self.register()
            if self.get_property("is-remote"):
                self.activate_action("quit")
            else:
                logging.warning("No running instance found")
            self.quit()
            return 0
        elif "new-note" in options:
            # Application needs to be registered to send action if remote instance
            self.register()
            self.activate_action("create-note")
        elif "open-note" in options:
            # Application needs to be registered to send action if remote instance
            self.register()
            note_id = GLib.Variant("u", options["open-note"])
            self.activate_action("open-note", note_id)
        elif "search" in options:
            # Application needs to be registered to send action if remote instance
            self.register()
            term = GLib.Variant("s", options["search"])
            self.activate_action("search-from-cli", term)
        elif "toggle-extended-preferences" in options:
            self.register()
            if self.__config_manager.show_extended_preferences:
                logging.info("Hiding extended preferences")
                self.__config_manager.show_extended_preferences = False
            else:
                logging.info("Showing extended preferences")
                self.__config_manager.show_extended_preferences = True
            self.quit()
            return 0

        # Let default option processing continue
        return -1

    def __on_activate(self) -> None:
        """Handle window activation."""
        if not self.__window:
            self.__create_window()
            assert self.__window  # mypy

        self.__apply_css()
        self.apply_style()
        self.__window.present()

    def __on_quit(self) -> None:
        if self.__window:
            self.__window.cleanup_and_close()

    def __on_create_note(self) -> None:
        self.activate()
        if self.__window:
            self.__window.activate_action("win.create-note-from-cli")

    def __on_create_note_from_uri(self, param: GObject.ParamSpec) -> None:
        self.activate()
        if self.__window:
            self.__window.activate_action("win.create-note-from-uri", param),

    def __on_open_note(self, param: GObject.ParamSpec) -> None:
        self.activate()
        if self.__window:
            self.__window.activate_action("win.open-note", param),

    def __on_search_from_cli(self, param: GObject.ParamSpec) -> None:
        self.activate()
        if self.__window:
            self.__window.activate_action("win.search-from-cli", param),

    def __add_cli_options(self) -> None:
        self.add_main_option(
            "new-note",
            ord("n"),
            GLib.OptionFlags.NONE,
            GLib.OptionArg.NONE,
            # Translators: Description, CLI option
            _("Create a note"),
            None,
        )
        self.add_main_option(
            "create-backup",
            0,
            GLib.OptionFlags.NONE,
            GLib.OptionArg.NONE,
            # Translators: Description, CLI option
            _("Create a backup"),
            None,
        )
        self.add_main_option(
            "restore-backup",
            0,
            GLib.OptionFlags.NONE,
            GLib.OptionArg.NONE,
            # Translators: Description, CLI option
            _("Restore a backup"),
            None,
        )
        self.add_main_option(
            "display-backup-path",
            0,
            GLib.OptionFlags.NONE,
            GLib.OptionArg.NONE,
            # Translators: Description, CLI option
            _("Display backup path"),
            None,
        )
        self.add_main_option(
            "display-ca-file-path",
            0,
            GLib.OptionFlags.NONE,
            GLib.OptionArg.NONE,
            # Translators: Description, CLI option
            _("Display path for custom server SSL CA chain file"),
            None,
        )
        self.add_main_option(
            "toggle-extended-preferences",
            0,
            GLib.OptionFlags.NONE,
            GLib.OptionArg.NONE,
            # Translators: Description, CLI option
            _("Toggle display of extended preferences in UI"),
            None,
        )
        self.add_main_option(
            "quit-running",
            0,
            GLib.OptionFlags.NONE,
            GLib.OptionArg.NONE,
            # Translators: Description, CLI option
            _("Quit any running instance"),
            None,
        )
        self.add_main_option(
            "debug-session",
            ord("d"),
            GLib.OptionFlags.NONE,
            GLib.OptionArg.NONE,
            # Translators: Description, CLI option
            _("Enable debug logging and functions"),
            None,
        )
        self.add_main_option(
            "open-note",
            0,
            GLib.OptionFlags.NONE,
            GLib.OptionArg.INT,
            # Translators: Description, CLI option
            _("Open note by id"),
            None,
        )
        self.add_main_option(
            "search",
            0,
            GLib.OptionFlags.NONE,
            GLib.OptionArg.STRING,
            # Translators: Description, CLI option
            _("Search in notes"),
            None,
        )

    def __setup_actions(self) -> None:
        def add_action(
            name: str,
            method: Callable,
            shortcut: Optional[str] = None,
            parameter_type: Optional[GLib.VariantType] = None,
        ):
            if parameter_type is None:
                action = Gio.SimpleAction.new(name)
            else:
                action = Gio.SimpleAction.new(name, parameter_type)
            self.add_action(action)
            action.connect("activate", method)
            self.__actions[name] = action
            if shortcut is not None:
                self.set_accels_for_action(f"app.{name}", [shortcut])

        add_action("about", lambda _o, _p: self.__show_about_dialog())
        add_action("quit", lambda _o, _p: self.__on_quit(), "<Control>q")
        add_action("settings", lambda _o, _p: self.__show_preferences_dialog(), "<Control>comma")
        add_action(
            "create-note",
            lambda _o, _p: self.__on_create_note(),
        )
        add_action(
            "create-note-from-uri",
            lambda _o, param: self.__on_create_note_from_uri(param),
            None,
            GLib.VariantType("s"),
        )
        add_action(
            "open-note",
            lambda _o, param: self.__on_open_note(param),
            None,
            GLib.VariantType("u"),
        )
        add_action(
            "search-from-cli",
            lambda _o, param: self.__on_search_from_cli(param),
            None,
            GLib.VariantType("s"),
        )
        add_action("create-backup", lambda _o, _p: self.__create_backup())

    def __create_window(self) -> None:
        self.__window = Window(self, self.db, self.__sync_manager)
        self.add_window(self.__window)

    def __apply_css(self) -> None:
        self.__apply_index_category_label_css()

    def __show_about_dialog(self) -> None:
        builder = Gtk.Builder()
        builder.add_from_resource("/org/gnome/World/Iotas/ui/about_dialog.ui")
        about_dialog = builder.get_object("about_dialog")
        about_dialog.set_application_icon(const.APP_ID)
        about_dialog.set_version(const.VERSION)
        about_dialog.set_developers(["Chris Heywood"])
        about_dialog.present(self.__window)

    def __show_preferences_dialog(self) -> None:
        if self.__preferences_dialog is not None:
            return

        self.__preferences_dialog = PreferencesDialog()
        self.__preferences_dialog.connect(
            "start-onboarding", self.__on_preferences_dialog_start_onboarding
        )
        self.__preferences_dialog.connect("closed", self.__on_preferences_dialog_closed)
        self.__preferences_dialog.present(self.__window)

    def __on_preferences_dialog_start_onboarding(self, dialog: PreferencesDialog) -> None:
        if not self.__preferences_dialog or not self.__window:  # mypy
            return
        self.__preferences_dialog.close()
        self.__window.show_login_dialog()

    def __on_preferences_dialog_closed(self, _dialog: GObject.Object) -> None:
        if not self.__preferences_dialog:  # mypy
            return
        self.__preferences_dialog.disconnect_by_func(self.__on_preferences_dialog_start_onboarding)
        self.__preferences_dialog = None

    def __create_backup(self) -> None:
        logging.info("Running remotely requested backup")
        self.__backup_manager.create_backup(
            self.__config_manager.nextcloud_sync_configured,
            self.__config_manager.backup_note_extension,
        )

    def __apply_index_category_label_css(self) -> None:
        # First pass approach to building this is using a display CSS provider, instead of adding
        # providers to each label. Presumably this is better, but it's a mess.
        style_manager = Adw.StyleManager.get_default()
        category_style = self.__config_manager.index_category_style
        colour_names = {
            "blue": {"light": "blue_2", "dark": "blue_5"},
            "green": {"light": "green_4", "dark": "green_5"},
            "yellow": {"light": "yellow_4", "dark": "yellow_5"},
            "orange": {"light": "orange_2", "dark": "orange_5"},
            "red": {"light": "red_1", "dark": "red_5"},
            "purple": {"light": "purple_2", "dark": "purple_5"},
        }
        if style_manager.get_dark():
            style_name = "dark"
        else:
            style_name = "light"

        css = "box label.index-category-pill {"
        if category_style in ("monochrome", "muted", "none"):
            css += """
              border-color: @insensitive_fg_color;
              border-width: 1px;
              border-style: solid;"""
        else:
            css += f"""
                background-color: @{colour_names[category_style][style_name]};
                color: @light_1;"""
        if category_style == "muted":
            # Matching libadwaita for dimmed/subtitle (non high-contrast, monochrome option is
            # provides for high contrast)
            css += "\nopacity: 0.55;"
        css += "\n}"

        display = Gdk.Display.get_default()
        if self.__index_category_label_css_provider is None:
            self.__index_category_label_css_provider = Gtk.CssProvider()
        else:
            Gtk.StyleContext.remove_provider_for_display(
                display, self.__index_category_label_css_provider
            )
        self.__index_category_label_css_provider.load_from_data(css, -1)
        Gtk.StyleContext.add_provider_for_display(
            display,
            self.__index_category_label_css_provider,
            Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION,
        )



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Gio

import html
from typing import Callable, NamedTuple, Optional

from markdown_it.token import Token
from markdown_it.utils import EnvType, OptionsDict

from iotas.html_generator import HtmlGenerator
from iotas.html_generator_configuration import HtmlGeneratorConfiguration
from iotas.note import Note
from iotas.text_utils import has_unicode


class TemplateDetails(NamedTuple):
    """Template string parameters."""

    html: str

    javascript: str

    css: str

    search_css: str


class TemplateHtmlGenerator(HtmlGenerator):
    """Generate HTML from note contents using a template."""

    EXPORT_CSS_PATH = "css/markdown.css"
    EXPORT_KATEX_CSS_PATH = "css/katex.min.css"
    EXPORT_KATEX_JS_PATH = "js/katex.min.js"

    TEXT_COLOUR_LIGHT = "#504e55"  # GtkSourceView Dark 3
    TEXT_COLOUR_LIGHT_HIGH_CONTRAST = "#000000"

    TEXT_COLOUR_DARK = "#c0bfbc"  # GtkSourceView Light 5
    TEXT_COLOUR_DARK_HIGH_CONTRAST = "#ffffff"

    CHECKED_TASK_COLOUR_LIGHT = "#b0afac"  # GtkSourceView Light 6
    CHECKED_TASK_COLOUR_LIGHT_HIGH_CONTRAST = "#000000"

    CHECKED_TASK_COLOUR_DARK = "#77767b"  # GtkSourceView Dark 1
    CHECKED_TASK_COLOUR_DARK_HIGH_CONTRAST = "#ffffff"

    def __init__(
        self,
        config_manager: HtmlGeneratorConfiguration,
        app_data_path: str,
        toolbar_underlay_padding_height: int,
        custom_template_loader: Optional[Callable[[], TemplateDetails]] = None,
    ) -> None:
        super().__init__()
        self.__config_manager = config_manager
        self.__app_data_path = app_data_path
        self.__toolbar_underlay_padding_height = toolbar_underlay_padding_height
        self.__font_family = ""
        self.__html_template: Optional[str] = None
        self.__javascript_template: str
        self.__configurable_css_template: str
        self.__searching_css: str
        self.__template_loader = (
            custom_template_loader
            if custom_template_loader
            else self.__populate_template_from_resources
        )

    def generate(
        self,
        note: Note,
        tokens: list[Token],
        render_func: Callable[[list[Token], OptionsDict, EnvType], str],
        parser_options: OptionsDict,
        searching: bool,
        export_format: Optional[str],
        scroll_position: Optional[float] = None,
    ) -> str:
        """Generator HTML for note.

        :param Note note: Note to render
        :param list[Token] tokens: Parser tokens
        :param Callable[[list[Token], OptionsDict, EnvType], str] render_fun: Render function
        :param OptionsDict parser_options: Parser options
        :param bool searching: Whether search CSS should be included
        :param Optional[str] export_format: Export format, if using
        :param Optional[float] scroll_position: Position to scroll to
        :return: Generated HTML
        :rtype: str
        """
        for token in tokens:
            if token.map:
                token.attrs["data-map"] = token.map[0]
        content = render_func(tokens, parser_options, None)

        if self.__html_template is None:
            self.__populate_templates()
        assert self.__html_template

        javascript = ""
        exporting = export_format is not None
        if not exporting:
            scroll_js = ""
            if scroll_position is not None:
                scroll_js = f"""document.documentElement.scrollTop = {scroll_position} *
                (document.documentElement.scrollHeight - document.documentElement.clientHeight);
                """
            javascript = self.__javascript_template % (str(not note.read_only).lower(), scroll_js)
            javascript = f"\n  <script>{javascript}</script>\n"

        if export_format == "html":
            css_path = self.EXPORT_CSS_PATH
            katex_css_path = self.EXPORT_KATEX_CSS_PATH
            katex_js_path = self.EXPORT_KATEX_JS_PATH
        else:
            css_path = f"{self.__app_data_path}/{self.RESOURCE_CSS_PATH}"
            katex_css_path = f"{self.__app_data_path}/{self.RESOURCE_KATEX_CSS_PATH}"
            katex_js_path = f"{self.__app_data_path}/{self.RESOURCE_KATEX_JS_PATH}"
        if self.__config_manager.markdown_tex_support:
            tex_headers = f'\n  <link rel="stylesheet" href="{katex_css_path}">'
            tex_headers += f'\n  <script src="{katex_js_path}"></script>'
        else:
            tex_headers = ""
        if has_unicode(content):
            meta_headers = '\n  <meta charset="utf-8">'
        else:
            meta_headers = ""

        configurable_stylesheet = ""
        if export_format == "html":
            stylesheet = self.generate_user_stylesheet(searching, -1)
            configurable_stylesheet = f"\n  <style>\n{stylesheet}  </style>"

        content = self.__html_template.format(
            title=html.escape(note.title),
            css_path=css_path,
            meta_headers=meta_headers,
            configurable_stylesheet=configurable_stylesheet,
            tex_headers=tex_headers,
            javascript=javascript,
            content=content,
        )
        return content

    def generate_user_stylesheet(self, searching: bool, height: int) -> str:
        """Generate part of stylesheet based on state (configuration etc).

        :param bool searching: Whether searching
        :param int height: View height
        :return: stylesheet
        :rtype: str
        """
        if self.__html_template is None:
            self.__populate_templates()

        font_size = self.__get_font_size()
        line_length = self.__config_manager.line_length
        backup_font = (
            "monospace" if self.__config_manager.markdown_use_monospace_font else "sans-serif"
        )
        families = [self.__font_family, backup_font]
        font_family = ", ".join(families)
        if self.__config_manager.editor_header_bar_visible_for_window_state:
            padding_top = 1
            thin_window_top_padding = "0"
        else:
            padding_top = self.__toolbar_underlay_padding_height
            thin_window_top_padding = "-webkit-calc(42px - 1em)"

        if self.__config_manager.high_contrast:
            light_text_colour = self.TEXT_COLOUR_LIGHT_HIGH_CONTRAST
            dark_text_colour = self.TEXT_COLOUR_DARK_HIGH_CONTRAST
            light_checked_task_colour = self.CHECKED_TASK_COLOUR_LIGHT_HIGH_CONTRAST
            dark_checked_task_colour = self.CHECKED_TASK_COLOUR_DARK_HIGH_CONTRAST
            figcaption_opacity = "1.0"
        else:
            light_text_colour = self.TEXT_COLOUR_LIGHT
            dark_text_colour = self.TEXT_COLOUR_DARK
            light_checked_task_colour = self.CHECKED_TASK_COLOUR_LIGHT
            dark_checked_task_colour = self.CHECKED_TASK_COLOUR_DARK
            figcaption_opacity = "0.85"

        if height > 0:
            padding_bottom = int(height * 0.4)
        else:
            padding_bottom = 0

        stylesheet = self.__configurable_css_template % (
            light_text_colour,
            figcaption_opacity,
            light_checked_task_colour,
            font_size,
            line_length,
            font_family,
            padding_top,
            padding_bottom,
            thin_window_top_padding,
            dark_text_colour,
            dark_checked_task_colour,
        )

        if searching:
            stylesheet += self.__searching_css

        return stylesheet

    def update_font_family(self, family: str) -> None:
        """Update the font family.

        :param str family: New font family
        """
        self.__font_family = family

    def __get_font_size(self) -> float:
        monospace_editor = self.__config_manager.use_monospace_font
        monospace_render = self.__config_manager.markdown_use_monospace_font
        editor_size = self.__config_manager.font_size
        ratio = self.__config_manager.markdown_render_monospace_font_ratio
        if monospace_editor and not monospace_render:
            size = editor_size / ratio
        elif monospace_render and not monospace_editor:
            size = editor_size * ratio
        else:
            size = editor_size
        return size

    def __populate_templates(self) -> None:
        if self.__html_template is not None:
            return
        details = self.__template_loader()
        self.__html_template = details.html
        self.__javascript_template = details.javascript
        self.__configurable_css_template = details.css
        self.__searching_css = details.search_css

    def __populate_template_from_resources(self) -> TemplateDetails:
        return TemplateDetails(
            self.__load_object_from_resource("template.html"),
            self.__load_object_from_resource("binding.js"),
            self.__load_object_from_resource("configurable.css"),
            self.__load_object_from_resource("searching.css"),
        )

    def __load_object_from_resource(self, filename: str) -> str:
        f = Gio.File.new_for_uri(f"resource:///org/gnome/World/Iotas/render-template/{filename}")
        status, contents, _tag = f.load_contents()
        if not status:
            raise Exception(f"Failed to load resource {filename}")

        return contents.decode("utf-8")



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Adw, GLib, Gtk

import logging

from iotas.config_manager import ConfigManager, HeaderBarVisibility
from iotas.ui_utils import have_window_with_width, is_likely_mobile_device


@Gtk.Template(resource_path="/org/gnome/World/Iotas/ui/first_start_page.ui")
class FirstStartPage(Adw.Bin):
    __gtype_name__ = "FirstStartPage"

    def initialise_device(self) -> None:
        """Apply any device-specific initialisation."""

        # Another location (ala MigrationAssistant) using this pattern as a result of not yet
        # discovering the clean way to be notified when a window has obtained its initial size
        def wait_and_run():
            if not have_window_with_width():
                GLib.timeout_add(250, wait_and_run)
            else:
                self.__set_formatting_bar_visibility_based_on_device()

        wait_and_run()

    def __set_formatting_bar_visibility_based_on_device(self) -> None:
        if is_likely_mobile_device():
            # Default setting is always visible so no change
            logging.info("Likely on mobile device, setting formatting bar to stay visible")
        else:
            logging.info("Likely not on mobile device, setting formatting bar to auto hide")
            ConfigManager.get_default().editor_formatting_bar_visibility = (
                HeaderBarVisibility.AUTO_HIDE
            )



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import GObject, GLib

import logging
import os
import queue
from threading import Thread
from typing import Callable, Optional


from iotas.attachment_helpers import (
    Attachment,
    get_attachment_disk_states,
)
from iotas.export_utils import (
    build_default_filename,
)
from iotas.markdown_helpers import (
    parse_to_tokens,
)
from iotas.note import Note
from iotas.note_exporter import NoteExporter


class BatchExporter(GObject.Object):
    """Batch note exporter.

    :param NoteExporter note_exporter: Single note exporter
    :param Callable[[Attachment], bool] download_attachment: Function to transfer an attachment from
        server, optional
    :param Callable[[Note], None] note_content_populator: Function to populate note content from
        storage
    """

    __gsignals__ = {
        # int: count
        "attachment-transfer-started": (GObject.SignalFlags.RUN_FIRST, None, (int,)),
        # float: progress
        "attachment-transferred": (GObject.SignalFlags.RUN_FIRST, None, (float,)),
        # int: count
        "finished-attachment-detection-transfer-run": (GObject.SignalFlags.RUN_FIRST, None, (int,)),
        "export-started": (GObject.SignalFlags.RUN_FIRST, None, ()),
        # float: progress
        "note-exported": (GObject.SignalFlags.RUN_FIRST, None, (float,)),
        # str: reason
        "export-failed": (GObject.SignalFlags.RUN_FIRST, None, (str,)),
        # str: single export out path
        "batch-finished": (GObject.SignalFlags.RUN_FIRST, None, (str,)),
    }

    def __init__(
        self,
        note_exporter: NoteExporter,
        download_attachment: Optional[Callable[[Attachment], bool]],
        note_content_populator: Optional[Callable[[Note], None]],
    ) -> None:
        super().__init__()
        self.__note_exporter = note_exporter
        self.__download_attachment = download_attachment
        self.__note_content_populator = note_content_populator
        self.__note_exporter.connect("finished", self.__on_single_finished)
        self.__note_exporter.connect("failed", self.__on_single_failed)

        self.__notes: list[Note] = []
        self.__location: Optional[str] = None
        self.__file_extension = ""
        self.__out_format = ""
        self.__supporting_tex = False

        self.__attachments_downloaded = 0
        self.__failed_attachment_transfers: set[int] = set()

        self.__export_queue: queue.Queue[Note] = queue.Queue()

    def check_and_transfer_missing_attachments(self) -> None:
        """Check for any missing attachments and transfer if necessary."""
        if not self.__download_attachment:
            raise Exception("Incorrect usage")

        def thread_do():
            self.__missing_attachments = {}

            for note in self.__notes:
                if not note.content_loaded:
                    self.__note_content_populator(note)

            for note in self.__notes:
                _parser, tokens = parse_to_tokens(note, exporting=False, tex_support=False)
                states = get_attachment_disk_states(note, tokens)
                self.__missing_attachments.update(states.missing)

            logging.debug(f"Have {len(self.__missing_attachments)} missing attachment/s")

            GLib.idle_add(self.__transfer_missing_attachments)

        thread = Thread(target=thread_do)
        thread.daemon = True
        thread.start()

    def start_export(self) -> None:
        """Kick off export."""
        assert self.__location  # mypy

        self.emit("export-started")

        if self.multi_export:
            try:
                os.makedirs(self.__location, exist_ok=True)
            except OSError as e:
                logging.warning(f"Failed to create directory: {e}")
                return

        self.__continue_export()

    def cleanup(self) -> None:
        """Flush export queue and disconnect signals."""
        self.__flush_export_queue()
        self.__note_exporter.disconnect_by_func(self.__on_single_finished)
        self.__note_exporter.disconnect_by_func(self.__on_single_failed)

    def get_notes(self) -> list[Note]:
        return self.__notes

    def set_notes(self, notes: list[Note]) -> None:
        self.__notes = notes
        self.__flush_export_queue()
        for note in self.__notes:
            self.__export_queue.put(note)

    @GObject.Property(type=str)
    def location(self) -> Optional[str]:
        return self.__location

    @location.setter
    def set_location(self, value: str) -> None:
        self.__location = value

    @GObject.Property(type=str)
    def out_format(self) -> str:
        return self.__out_format

    @out_format.setter
    def set_out_format(self, value: str) -> None:
        self.__out_format = value

    @GObject.Property(type=str)
    def file_extension(self) -> str:
        return self.__file_extension

    @file_extension.setter
    def set_file_extension(self, value: str) -> None:
        self.__file_extension = value

    @GObject.Property(type=bool, default=False)
    def supporting_tex(self) -> bool:
        return self.__supporting_tex

    @supporting_tex.setter
    def set_supporting_tex(self, value: bool) -> None:
        self.__supporting_tex = value

    @GObject.Property(type=bool, default=False)
    def multi_export(self) -> bool:
        return len(self.__notes) > 1

    def __on_single_finished(self, _obj: GObject.Object, path: str) -> None:
        remaining = self.__export_queue.qsize()
        progress = (len(self.__notes) - remaining) / len(self.__notes)
        self.emit("note-exported", progress)

        if not self.__continue_export():
            if self.multi_export:
                path = os.path.abspath(os.path.join(path, os.pardir))
            self.emit("batch-finished", path)

    def __on_single_failed(self, _obj: GObject.Object, reason: str) -> None:
        self.emit("export-failed", reason)

    def __transfer_missing_attachments(self) -> None:
        count = len(self.__missing_attachments)
        if count == 0:
            self.emit("finished-attachment-detection-transfer-run", 0)
            return

        self.__attachments_downloaded = 0

        logging.info("Downloading attachments prior to export")
        self.emit("attachment-transfer-started", count)

        self.__failed_attachment_transfers = set()

        def thread_do():
            failures = []

            for attachment in self.__missing_attachments.values():
                if self.__download_attachment(attachment):
                    self.__attachments_downloaded += 1
                    progress = self.__attachments_downloaded / len(self.__missing_attachments)
                    GLib.idle_add(self.emit, "attachment-transferred", progress)
                else:
                    failures.append(attachment)

            self.__failed_attachment_transfers = {attachment.note_id for attachment in failures}
            GLib.idle_add(self.emit, "finished-attachment-detection-transfer-run", len(failures))

        thread = Thread(target=thread_do)
        thread.daemon = True
        thread.start()

    def __continue_export(self) -> bool:
        try:
            note = self.__export_queue.get_nowait()
        except queue.Empty:
            return False

        missing_attachments = note.id in self.__failed_attachment_transfers
        location: str | None
        if self.multi_export:
            assert self.__location  # We ensure location is set for multi-export

            # For the local backend note contents are populated here instead of during attachment
            # download checking. The contents need to be populated prior to filename building.
            self.__ensure_content_loaded(note)

            filename = build_default_filename(note, self.__out_format, self.__file_extension)
            location = os.path.join(self.__location, filename)
        else:
            location = self.__location

        self.__note_exporter.export(
            note,
            self.__out_format,
            self.__file_extension,
            self.__supporting_tex,
            missing_attachments,
            location,
        )
        return True

    def __ensure_content_loaded(self, note: Note) -> None:
        if not note.content_loaded:
            # The populator won't be provided if exporting from the editor, but in that case
            # the content will always be populated
            assert self.__note_content_populator
            self.__note_content_populator(note)

    def __flush_export_queue(self) -> None:
        while not self.__export_queue.empty():
            try:
                self.__export_queue.get_nowait()
            except queue.Empty:
                pass



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import GObject, Gtk, GtkSource

from enum import IntEnum, auto
import logging
import math
import re
from typing import Callable, Optional

from iotas.ordered_list_utils import (
    check_line_for_ordered_list_item,
    calculate_ordered_list_index,
    format_ordered_list_item,
)
from iotas.text_utils import (
    get_iters_at_sourceview_context_class_extents,
    iter_forward_to_context_class_removed,
    parse_any_url_at_iter,
    parse_markdown_automatic_link,
    parse_markdown_inline_link,
    str_is_url,
)


class LinkStateInfo(GObject.Object):
    """State information for a markup link in the buffer."""

    creating: bool = True
    """Creating a new link"""

    url_angle_bracketed: bool = False
    """Editing link with URL angle-bracket wrapped"""

    automatic_link: bool = False
    """Creating a new link from an automatic link"""

    text: str = ""
    """Link text"""

    link: str = ""
    """Link URL"""

    title: str = ""
    """Link title, which gets stored and recreated"""

    source_text: str = ""
    """Source text for a link being edited. Used for verification if a note is server updated."""

    start_offset: int = -1
    """Start offset for a link being edited. Used for verification if a note is server updated."""

    end_offset: int = -1
    """End offset for a link being edited. Used for verification if a note is server updated."""


class _ListType(IntEnum):
    """Markup list type."""

    UNORDERED = auto()
    ORDERED = auto()
    TASK = auto()


class Formatter(GObject.Object):
    """Markup formatter."""

    def __init__(self) -> None:
        super().__init__()
        self.__buffer: GtkSource.Buffer

    def bold(self) -> None:
        """Apply user bold formatting action.

        This may result in adding or removing formatting, depending on cursor and selection state,
        note contents, etc.
        """
        insert = self.__get_selection_start_or_insert_iter()
        if "bold" in self.__buffer.get_context_classes_at_iter(insert):
            # In bold, remove
            self.__delete_wrapping_markup("bold", insert, 2)
        elif self.__buffer.get_has_selection() and not self.__have_multi_line_selection():
            # Have single line selection, wrap in bold
            self.__wrap_selection_in_markup("*", 2)
        else:
            text = self.__buffer.get_property("text")
            if self.__in_empty_bold(text, insert.get_offset()):
                # Handle toggle
                self.__delete_chars_around_iter(insert, 2)
            else:
                # Insert empty bold markup around cursor
                self.__insert_and_centre_iter(insert, "****")

    def italic(self) -> None:
        """Apply user italic formatting action.

        This may result in adding or removing formatting, depending on cursor and selection state,
        note contents, etc.
        """
        insert = self.__get_selection_start_or_insert_iter()
        if "italic" in self.__buffer.get_context_classes_at_iter(insert):
            # In italic, remove
            self.__delete_wrapping_markup("italic", insert, 1)
        elif self.__buffer.get_has_selection() and not self.__have_multi_line_selection():
            # Have single line selection, wrap in italic
            self.__wrap_selection_in_markup("_", 1)
        else:
            text = self.__buffer.get_property("text")
            if self.__in_empty_italics(text, insert.get_offset()):
                # Handle toggle
                self.__delete_chars_around_iter(insert, 1)
            else:
                # Insert empty italic markup around cursor
                self.__insert_and_centre_iter(insert, "__")

    def strikethrough(self) -> None:
        """Apply user strikethrough formatting action.

        This may result in adding or removing formatting, depending on cursor and selection state,
        note contents, etc.
        """
        insert = self.__get_selection_start_or_insert_iter()
        if "strikethrough" in self.__buffer.get_context_classes_at_iter(insert):
            # In strikethrough, remove
            self.__delete_wrapping_markup("strikethrough", insert, 2)
        elif self.__buffer.get_has_selection() and not self.__have_multi_line_selection():
            # Have single line selection, wrap in strikethrough
            self.__wrap_selection_in_markup("~", 2)
        else:
            text = self.__buffer.get_property("text")
            if self.__in_empty_strikethrough(text, insert.get_offset()):
                # Handle toggle
                self.__delete_chars_around_iter(insert, 2)
            else:
                # Insert empty strikethrough markup around cursor
                self.__insert_and_centre_iter(insert, "~~~~")

    def quote(self) -> None:
        """Apply user quote formatting action.

        This may result in adding or removing formatting, depending on cursor and selection state,
        note contents, etc.
        """
        insert = self.__get_selection_start_or_insert_iter()
        line_start = insert.copy()
        line_start.set_line_offset(0)

        if self.__have_multi_line_selection():
            # Handle multi-line selection
            minimum_depth = self.__get_minimum_quote_marker_depth_across_selection_lines()
            if minimum_depth == 0:
                # Some lines have no marker, add markers to all
                logging.debug("Adding quote level")
                self.__run_method_on_selection_lines(self.__add_quote_marker_to_line)
            else:
                # All lines have a marker, remove a level
                logging.debug("Removing quote level")
                self.__run_method_on_selection_lines(self.__remove_quote_marker_from_line)
        elif "blockquote" in self.__buffer.get_context_classes_at_iter(line_start):
            # In quote, remove
            logging.debug("Removing quote")
            self.__buffer.begin_user_action()
            self.__remove_quote_marker_from_line(line_start)
            self.__buffer.end_user_action()
        else:
            # Prepend line with quote markup
            logging.debug("Adding quote")
            self.__buffer.begin_user_action()
            self.__add_quote_marker_to_line(line_start)
            self.__buffer.end_user_action()

    def heading(self, new_level: int) -> None:
        """Apply user heading formatting action.

        This may result in adding or removing formatting, depending on cursor and selection state,
        note contents, requested level, etc.

        :param int new_level: Requested header level, zero to remove
        """
        insert = self.__get_selection_start_or_insert_iter()
        line_start = insert.copy()
        line_start.set_line_offset(0)

        # Check for existing heading
        heading_start = line_start.copy()
        existing_level = 0
        while not heading_start.ends_line():
            char = self.__get_line_char_at_iter(heading_start)
            if char == "#":
                heading_counter = heading_start.copy()
                while self.__get_line_char_at_iter(heading_counter) == "#":
                    existing_level += 1
                    heading_counter.forward_char()

                break
            heading_start.forward_char()

        if existing_level == new_level:
            logging.info(f"Nothing to do, already have heading level {new_level}")
            return

        self.__buffer.begin_user_action()

        # Remove any existing heading
        if existing_level:
            end = heading_start.copy()
            end.forward_chars(existing_level)
            char = self.__get_line_char_at_iter(end)
            if char.isspace():
                end.forward_char()
            self.__buffer.delete(heading_start, end)
        else:
            heading_start = line_start.copy()

        # Insert new heading
        if new_level:
            self.__buffer.insert(heading_start, "#" * new_level)
            self.__insert_space_if_not_exists(heading_start)

        self.__buffer.end_user_action()

    def get_link_state(self) -> LinkStateInfo:
        """Fetch link state information for context.

        Returned object will vary depending on note contents, cursor location or any selection.

        :return: Link information at cursor/selection
        :rtype: LinkStateInfo
        """
        info = LinkStateInfo()
        mark = self.__buffer.get_insert()
        if self.__buffer.get_has_selection():
            # Use selection
            start, end = self.__buffer.get_selection_bounds()
            selection_start_classes = self.__buffer.get_context_classes_at_iter(start)
            text = self.__buffer.get_text(start, end, include_hidden_chars=False)
            if "\n" in text:
                # Multiline selection, ignore selection and create new link
                logging.debug("Have newline, ignoring selection")
                info.creating = True
            elif "inline-link" in selection_start_classes:
                # Start of selection in existing inline link, parse and edit
                if self.__parse_link(start, info):
                    info.creating = False
                else:
                    # Failed to parse, ignore selection and create new link
                    info.creating = True
            elif "automatic-link" in selection_start_classes:
                # Start of selection in automatic link, parse and edit, converting to inline link
                logging.debug("Have automatic-link")
                info.creating = False
                self.__parse_automatic_link(start, info)
            elif self.__selection_has_context_classes("bare-link"):
                # There are some (language parsing) context classes in the selection, let's just
                # ignore the selection and create new link
                logging.debug("Not using selection for URL text as it appears to have formatting")
                info.creating = True
            elif str_is_url(text, http_only=False):
                # Selected text is URL, populate into link creation dialog
                info.creating = True
                info.link = text
                info.source_text = text
                info.start_offset = start.get_offset()
                info.end_offset = end.get_offset()
                return info
            elif self.__link_scheme_in_text(text):
                # There's a URI scheme somewhere in the selection but we haven't managed to
                # identify the whole selection as a URL. Ignore the selection and create a new link.
                info.creating = True
            else:
                # Populate the selected text into link creation dialog
                logging.debug("Using selection text")
                info.creating = True
                info.text = text
                info.source_text = text
                info.start_offset = start.get_offset()
                info.end_offset = end.get_offset()
        else:
            # No selection
            insert = self.__buffer.get_iter_at_mark(mark)
            classes = self.__buffer.get_context_classes_at_iter(insert)
            if "inline-link" in classes:
                # Existing inline link under cursor, parse and edit
                if self.__parse_link(insert, info):
                    info.creating = False
                else:
                    # Failed to parse, ignore link under cursor and create new link
                    info.creating = True
            elif "automatic-link" in classes:
                # Cursor in automatic link, parse and edit, converting to inline link
                logging.debug("Have automatic-link")
                info.creating = True
                self.__parse_automatic_link(insert, info)
            elif self.__check_in_url_and_populate_state(insert, info):
                # Cursor in middle of URL, populate to link creation dialog
                info.creating = True
            else:
                # Create new link
                info.creating = True
        return info

    def link(self, info: LinkStateInfo) -> None:
        """Apply user link formatting action.

        This applies the result coming out of the link dialog.

        :param LinkStateInfo info: Details on the changes to be made
        """
        self.__buffer.begin_user_action()
        if self.__buffer.get_has_selection():
            insert, _ = self.__buffer.get_selection_bounds()
            insert_mark = self.__buffer.create_mark(None, insert)
            cleanup_mark = True
        else:
            insert_mark = self.__buffer.get_insert()
            cleanup_mark = False

        # If the edit is based on existing buffer text, delete that text after ensuring it hasn't
        # been modified by a remote update
        if info.source_text != "":
            start = self.__buffer.get_iter_at_offset(info.start_offset)
            end = self.__buffer.get_iter_at_offset(info.end_offset)
            text = self.__buffer.get_text(start, end, include_hidden_chars=False)
            if info.source_text != text:
                logging.warning("Note appears to have changed, inserting link instead")
            else:
                self.__buffer.delete(start, end)

        insert = self.__buffer.get_iter_at_mark(insert_mark)
        self.__insert_link(insert, info)
        # Nudge cursor back into link to enable immediate re-edit
        insert.backward_char()
        self.__buffer.place_cursor(insert)
        if cleanup_mark:
            self.__buffer.delete_mark(insert_mark)
        self.__buffer.end_user_action()

    def code(self) -> None:
        """Apply user code formatting action.

        This may result in adding or removing formatting, depending on cursor and selection state,
        note contents, etc.
        """
        insert = self.__get_selection_start_or_insert_iter()
        context_classes = self.__buffer.get_context_classes_at_iter(insert)
        if "code-block" in context_classes:
            # In code block, remove
            self.__buffer.begin_user_action()
            self.__delete_wrapping_markup("code-block", insert, 3, user_action=False)

            # Also remove the newline inserted on the starting markup, useful quick toggle to undo
            insert = self.__get_selection_start_or_insert_iter()
            prev_char = insert.copy()
            prev_char.backward_char()
            self.__buffer.delete(prev_char, insert)

            self.__buffer.end_user_action()
        elif "code-span" in context_classes:
            # In code span, remove
            backticks = 2 if "2-backticks-code-span" in context_classes else 1
            self.__delete_wrapping_markup("code-span", insert, backticks)
        elif not self.__buffer.get_has_selection() or "indented-code-block" in context_classes:
            # Have no selection, or have indented code block
            # For now indented codeblocks are basically ignored
            if "indented-code-block" in context_classes:
                logging.info(
                    "Your indented codeblock likely isn't being handled as you'd hope, sorry!"
                )
            text = self.__buffer.get_property("text")
            if self.__in_empty_codespan(text, insert.get_offset()):
                # Handle toggle
                self.__delete_chars_around_iter(insert, 1)
            else:
                # Insert code span markup around cursor
                self.__insert_and_centre_iter(insert, "``")
        else:
            # Have selection
            start, end = self.__buffer.get_selection_bounds()
            start_mark = self.__buffer.create_mark(None, start)
            end_mark = self.__buffer.create_mark(None, end)
            if start.get_line() != end.get_line():
                # Multi line selection, add code block markup
                logging.debug("Adding code-block for multi-line selection")
                chars = "```\n"
                # If selection doesn't start a line insert a newline before the start markup
                if not start.starts_line():
                    chars = "\n" + chars
                self.__buffer.begin_user_action()
                self.__buffer.insert(start, chars)
                end = self.__buffer.get_iter_at_mark(end_mark)
                chars = "```"

                if not end.starts_line():
                    # Selection doesn't end a line insert a newline before the end markup
                    chars = "\n" + chars + "\n"
                elif not self.__blank_line_follows(end):
                    # Not following blank line, single newline after
                    chars += "\n"
                self.__buffer.insert(end, chars)
                self.__buffer.end_user_action()

                # Retain selection
                start = self.__buffer.get_iter_at_mark(start_mark)
                end.backward_chars(len(chars))
                self.__buffer.select_range(start, end)
            else:
                # Single line selection, add code span markup
                logging.debug("Adding code-span for single-line selection")
                self.__buffer.begin_user_action()
                self.__buffer.insert(start, "`")
                end = self.__buffer.get_iter_at_mark(end_mark)
                self.__buffer.insert(end, "`")

                # Retain selection
                start = self.__buffer.get_iter_at_mark(start_mark)
                end = self.__buffer.get_iter_at_mark(end_mark)
                end.backward_chars(1)
                self.__buffer.select_range(start, end)
                # Cleanup
                self.__buffer.delete_mark(start_mark)
                self.__buffer.delete_mark(end_mark)
                self.__buffer.end_user_action()

    def unordered_list(self) -> None:
        """Apply user unordered list formatting action.

        This may result in adding or removing formatting, depending on cursor and selection state,
        note contents, etc.
        """
        logging.debug("unordered-list")
        self.__process_list(_ListType.UNORDERED, lambda iter: self.__buffer.insert(iter, "-"))

    def ordered_list(self) -> None:
        """Apply user ordered list formatting action.

        This may result in adding or removing formatting, depending on cursor and selection state,
        note contents, etc.
        """
        logging.debug("ordered-list")

        # If the previous line contains an ordered list item attempt to continue that list
        def add_marker(iter: Gtk.TextIter) -> None:
            # Default
            value = "1."

            line_start = iter.copy()
            line_start.set_line_offset(0)
            if not line_start.is_start():
                # Get previous line text
                previous_line_start = line_start.copy()
                previous_line_start.backward_line()
                line_end = previous_line_start.copy()
                line_end.forward_to_line_end()
                previous_line = self.__buffer.get_text(
                    previous_line_start, line_end, include_hidden_chars=False
                )
                # Check if it contains an ordered list item
                regex_match = check_line_for_ordered_list_item(previous_line)
                if regex_match:
                    _, index, marker = regex_match.groups()
                    spacing = ""
                    # Determine the value of the next item in the sequence
                    sequence_next = calculate_ordered_list_index(index, 1)
                    if sequence_next:
                        value = format_ordered_list_item(spacing, sequence_next, marker).strip()

            self.__buffer.insert(iter, value)

        self.__process_list(_ListType.ORDERED, add_marker)

    def checkbox(self) -> None:
        """Apply user checkbox formatting action.

        This may result in adding or removing formatting, depending on cursor and selection state,
        note contents, etc.
        """
        logging.debug("checkbox")
        self.__process_list(_ListType.TASK, lambda iter: self.__buffer.insert(iter, "- [ ]"))

    def toggle_checkbox(self) -> None:
        """Toggle any checkbox on the current line or selected lines."""
        logging.debug("toggle-checkbox")
        self.__buffer.begin_user_action()

        if self.__buffer.get_has_selection():
            # Have selection, perform toggle for each line in selection
            self.__run_method_on_selection_lines(self.__toggle_checkbox_single_line)
        else:
            # No selection, toggle any checkpoint on cursor line
            mark = self.__buffer.get_insert()
            insert = self.__buffer.get_iter_at_mark(mark)
            line_start = insert.copy()
            line_start.set_line_offset(0)
            self.__toggle_checkbox_single_line(line_start)

        self.__buffer.end_user_action()

    def horizontal_rule(self) -> None:
        """Apply user horizontal rule formatting action.

        This may result in adding or removing formatting, depending on cursor and selection state,
        note contents, etc.
        """
        insert = self.__get_selection_start_or_insert_iter()
        line_start = insert.copy()
        line_start.set_line_offset(0)

        # Check for existing horizontal rule on the cursor line, or the line at the start of the
        # selection
        start = line_start.copy()
        found = False
        while not start.ends_line():
            if "horizontal-rule" in self.__buffer.get_context_classes_at_iter(start):
                found = True
                break
            start.forward_char()

        if found:
            # Horizontal rule found on line, remove
            logging.debug("Removing horizontal rule")
            end = line_start.copy()
            if not iter_forward_to_context_class_removed(end, "horizontal-rule"):
                return
            self.__buffer.delete(start, end)
        else:
            # Add horizontal rule markup
            logging.debug("Adding horizontal rule")
            chars = "- - -"
            if not insert.starts_line():
                # Not at start of line, double newline before
                chars = "\n\n" + chars
            elif not self.__after_blank_line(insert):
                # Not preceding blank line, single newline before
                chars = "\n" + chars

            # Track suffixed chars separately so we can place the cursor back in the markup to
            # allow eg. toggling
            suffix_chars = ""
            if not insert.ends_line():
                # Not at end of line, double newline after
                suffix_chars = "\n\n"
            elif not self.__blank_line_follows(insert):
                # Not following blank line, single newline after
                suffix_chars = "\n"

            self.__buffer.insert(insert, chars + suffix_chars)
            insert.backward_chars(len(suffix_chars))
            self.__buffer.place_cursor(insert)

    def table(self, columns: int, rows: int, view_width: int, character_width: float) -> None:
        """Insert a table.

        :param int columns: Number of columns
        :param int rows: Number of rows (below header)
        :param int view_width: View pixel width
        :param float character_width: Estimated pixel width of single character in font
        """
        logging.info(f"Insert {columns}x{rows} table")
        insert = self.__get_selection_start_or_insert_iter()

        # Attempt to insert a valid table which doesn't wrap for the current view width.
        # First, cap table pixel width for wide views
        view_width = min(view_width, 1000)
        character_width = max(1.0, character_width)
        # Calculate max number of characters for the width
        max_chars = math.floor(view_width / character_width * 0.9)
        per_column_internal_chars = int((max_chars - 1 - columns) / columns)

        # Minimum 3 characters column width, by GFM standard
        column_width = max(3, per_column_internal_chars)
        # Cap column width
        column_width = min(per_column_internal_chars, 20)

        table = ""
        for row in range(rows + 1):
            line = "|"
            for _ in range(columns):
                line += " " * column_width + "|"
            table += line + "\n"
            if row == 0:
                # Add header separator
                line = "|"
                for _ in range(columns):
                    line += "-" * column_width + "|"
                table += line + "\n"
        self.__buffer.insert(insert, table)

    @GObject.Property(type=GtkSource.Buffer)
    def buffer(self) -> GtkSource.Buffer:
        return self.__buffer

    @buffer.setter
    def set_buffer(self, value: GtkSource.Buffer) -> None:
        self.__buffer = value

    def __get_wrapping_characters(
        self, start: Gtk.TextIter, end: Gtk.TextIter, characters: int
    ) -> Optional[str]:
        text = self.__buffer.get_property("text")
        if start.get_offset() < characters or end.get_offset() > len(text) - characters:
            # Handling bad request, being at start or end of buffer
            return None
        return text[start.get_offset() - characters : end.get_offset() + characters]

    def __get_line_char_at_iter(self, location: Gtk.TextIter) -> str:
        if location.ends_line():
            return ""
        else:
            return self.__get_char_at_iter(location)

    def __get_char_at_iter(self, location: Gtk.TextIter) -> str:
        char_end = location.copy()
        char_end.forward_char()
        return self.__buffer.get_text(location, char_end, include_hidden_chars=False)

    def __insert_space_if_not_exists(self, location: Gtk.TextIter) -> None:
        if location.ends_line():
            self.__buffer.insert(location, " ")
        else:
            char = self.__get_line_char_at_iter(location)
            if not char.isspace():
                self.__buffer.insert(location, " ")

    def __in_empty_bold(self, text: str, location: int) -> bool:
        markup_width = 2
        return self.__check_if_expression_occurs_at_location(
            text,
            r"(?<![*_])[*_]{4}(?![*_])",
            location - markup_width,
        )

    def __in_empty_italics(self, text: str, location: int) -> bool:
        markup_width = 1
        return self.__check_if_expression_occurs_at_location(
            text,
            r"(?<![_])[_]{2}(?![_])",
            location - markup_width,
        )

    def __in_empty_strikethrough(self, text: str, location: int) -> bool:
        markup_width = 2
        return self.__check_if_expression_occurs_at_location(
            text,
            r"(?<![~])[~]{4}(?![~])",
            location - markup_width,
        )

    def __in_empty_codespan(self, text: str, location: int) -> bool:
        markup_width = 1
        return self.__check_if_expression_occurs_at_location(
            text,
            r"(?<![`])[`]{2}(?![`])",
            location - markup_width,
        )

    def __check_if_expression_occurs_at_location(
        self, text: str, regex: str, location: int
    ) -> bool:
        """Check if regular expression occurs in text at specified location.

        :param str text: Text to search
        :param str regex: Regex
        :param int location: Location as offset
        :return: Whether matched
        :rtype: bool
        """
        success = False
        match_iter = re.finditer(regex, text)
        for match in match_iter:
            if match.start() == location:
                success = True
                break
        return success

    def __parse_link(self, inside: Gtk.TextIter, info: LinkStateInfo) -> bool:
        """Parse link elements from a known inline link.

        :param Gtk.TextIter inside: Iter inside inline link
        :param LinkStateInfo: Object to populate into
        :return: Success
        :rtype: bool
        """
        result = parse_markdown_inline_link(inside)
        if not result:
            return False

        info.link = result.link
        info.text = result.text
        info.source_text = result.source_text
        info.start_offset = result.start_offset
        info.end_offset = result.end_offset
        info.url_angle_bracketed = result.url_angle_bracketed

        return True

    def __parse_automatic_link(self, inside: Gtk.TextIter, info: LinkStateInfo) -> bool:
        """Parse URL from a known automatic link.

        :param Gtk.TextIter inside: Iter inside inline link
        :param LinkStateInfo: Object to populate into
        :return: Success
        :rtype: bool
        """
        info.creating = True
        result = parse_markdown_automatic_link(inside)
        if result:
            info.automatic_link = True
            info.title = ""
            info.link = result.link
            info.source_text = result.source_text
            info.start_offset = result.start_offset
            info.end_offset = result.end_offset
            return True
        else:
            # Failure very unlikely as we depend on the GtkSourceView language parsing to reach
            # here
            logging.warning("Failed to parse automatic-link")
            return False

    def __delete_wrapping_markup(
        self, cls: str, inside: Gtk.TextIter, width: int, user_action: bool = True
    ) -> None:
        logging.debug(f"Removing wrapping '{cls}'")
        start, end = get_iters_at_sourceview_context_class_extents(cls, inside)
        if not start:
            return
        end_mark = self.__buffer.create_mark(None, end)
        tmp = start.copy()
        tmp.forward_chars(width)
        if user_action:
            self.__buffer.begin_user_action()
        self.__buffer.delete(start, tmp)
        end = self.__buffer.get_iter_at_mark(end_mark)
        tmp = end.copy()
        end.backward_chars(width)
        self.__buffer.delete(end, tmp)
        self.__buffer.delete_mark(end_mark)
        if user_action:
            self.__buffer.end_user_action()

    def __wrap_selection_in_markup(self, char: str, count: int) -> None:
        """Wrap selection in markup.

        :param str char: Character to insert
        :param int count: Number of times to insert the character on each side of the selection
        """
        logging.debug(f"Wrapping selection with '{char*count}'")
        start, end = self.__buffer.get_selection_bounds()
        start_mark = self.__buffer.create_mark(None, start)
        end_mark = self.__buffer.create_mark(None, end)
        self.__buffer.begin_user_action()
        self.__buffer.insert(start, char * count)
        end = self.__buffer.get_iter_at_mark(end_mark)
        self.__buffer.insert(end, char * count)

        # Retain selection
        start = self.__buffer.get_iter_at_mark(start_mark)
        end = self.__buffer.get_iter_at_mark(end_mark)
        end.backward_chars(count)
        self.__buffer.select_range(start, end)
        # Cleanup
        self.__buffer.delete_mark(start_mark)
        self.__buffer.delete_mark(end_mark)
        self.__buffer.end_user_action()

    def __delete_chars_around_iter(self, inside: Gtk.TextIter, either_side_count: int) -> None:
        start = inside.copy()
        start.backward_chars(either_side_count)
        end = start.copy()
        end.forward_chars(either_side_count * 2)
        self.__buffer.delete(start, end)

    def __insert_and_centre_iter(self, location: Gtk.TextIter, value: str) -> None:
        """Insert text and centre cursor in new text.

        :param Gtk.TextIter location: Location to insert
        :param str value: Text to insert
        """
        if value == "":
            return
        insert = location.copy()
        mark = self.__buffer.create_mark(None, location)
        self.__buffer.insert(insert, value)
        insert = self.__buffer.get_iter_at_mark(mark)
        insert.backward_chars(int(len(value) / 2))
        self.__buffer.place_cursor(insert)
        self.__buffer.delete_mark(mark)

    def __get_selection_start_or_insert_iter(self) -> Gtk.TextIter:
        """Fetch an iter at either the start of the selection or cursor.

        :return: The iter
        :rtype: Gtk.TextIter
        """
        if self.__buffer.get_has_selection():
            insert, _ = self.__buffer.get_selection_bounds()
        else:
            insert = self.__buffer.get_iter_at_mark(self.__buffer.get_insert())
        return insert

    def __selection_has_context_classes(self, ignore: str) -> bool:
        """Check whether any context class exists in the selection.

        :param str ignore: Class to ignore
        :return: Whether any context class exists
        :rtype: bool
        """
        if not self.__buffer.get_has_selection():
            logging.warning("No selection?")
            return False

        have_classes = False
        start, end = self.__buffer.get_selection_bounds()
        loc = start.copy()
        while loc.get_offset() != end.get_offset():
            classes = self.__buffer.get_context_classes_at_iter(loc)
            only_ignored_class = len(classes) == 1 and classes[0] == ignore
            have_classes = len(classes) > 0 and not only_ignored_class
            if have_classes:
                break
            loc.forward_char()

        return have_classes

    def __check_in_url_and_populate_state(self, iter: Gtk.TextIter, info: LinkStateInfo) -> bool:
        """Check whether location in buffer is in a URL and if so populate it into link info.

        :param Gtk.TextIter iter: Iter
        :param LinkStateInfo: Object to populate into
        :return: Whether URL found
        :rtype: bool
        """
        result = parse_any_url_at_iter(iter, http_only=False)
        if result:
            info.link = result.link
            info.source_text = result.link
            info.start_offset = result.start_offset
            info.end_offset = result.end_offset
            return True
        else:
            return False

    def __link_scheme_in_text(self, text: str) -> bool:
        return "https://" in text or "http://" in text

    def __insert_link(self, insert: Gtk.TextIter, info: LinkStateInfo) -> None:
        link = info.link
        if " " in link or info.url_angle_bracketed:
            link = f"<{link}>"
        title = ""
        if info.title.strip() != "":
            title = f' "{info.title}"'
        self.__buffer.insert(insert, f"[{info.text}]({link}{title})")

    def __process_list(self, list_type: _ListType, add_marker: Callable) -> None:
        """Apply user list action to all lines in a selection or on the cursor line.

        :param _ListType list_type: Type of list the user has actioned
        :param Callable add_mark: Method to call to add marker
        """
        self.__buffer.begin_user_action()

        if self.__buffer.get_has_selection():

            def callback(line_start: Gtk.TextIter):
                self.__single_line_list_marker_action(line_start, list_type, add_marker)

            self.__run_method_on_selection_lines(callback)
        else:
            mark = self.__buffer.get_insert()
            insert = self.__buffer.get_iter_at_mark(mark)
            line_start = insert.copy()
            line_start.set_line_offset(0)
            self.__single_line_list_marker_action(line_start, list_type, add_marker)

        self.__buffer.end_user_action()

    def __run_method_on_selection_lines(self, method: Callable) -> None:
        if not self.__buffer.get_has_selection():
            logging.warning("No selection, incorrect method use")
            return

        insert, end = self.__buffer.get_selection_bounds()
        end_mark = self.__buffer.create_mark(None, end)
        line_start = insert.copy()
        line_start.set_line_offset(0)
        while line_start.compare(end) < 0 and not line_start.is_end():
            insert_mark = self.__buffer.create_mark(None, line_start)
            method(line_start)
            line_start = self.__buffer.get_iter_at_mark(insert_mark)
            line_start.forward_line()
            end = self.__buffer.get_iter_at_mark(end_mark)
            self.__buffer.delete_mark(insert_mark)
        self.__buffer.delete_mark(end_mark)

    def __single_line_list_marker_action(
        self, line_start: Gtk.TextIter, list_type: _ListType, add_marker: Callable
    ) -> None:
        # Set GtkSourceView context class from list type
        cls = "checkbox" if list_type == _ListType.TASK else "list-marker"
        # Build list of the types, excuding our actioned type
        other_types = list(_ListType)
        other_types.remove(list_type)

        # Check for and remove  actioned type
        removed = False
        if cls in self.__buffer.get_context_classes_at_iter(line_start):
            removed = self.__remove_existing_list_marker_on_line(line_start, [list_type])

        # If we haven't removed our actioned marker type from the line add it
        if not removed:
            logging.debug(f"No {list_type.name.lower()} marker removed, adding")
            self.__remove_existing_list_marker_on_line(line_start, other_types)
            add_marker(line_start)
            self.__insert_space_if_not_exists(line_start)

    def __remove_existing_list_marker_on_line(
        self, insert: Gtk.TextIter, types: list[_ListType]
    ) -> bool:
        """Remove any existing list marker for provided list types on line.

        :param Gtk.TextIter: Iter on line
        :param list[_ListType]: Types to remove
        :return: Whether any marker was removed
        :rtype: bool
        """
        if len(types) == 0:
            logging.warning("No list type to remove")
            return False

        # Build the GtkSourceView context classes to remove
        remove_classes = []
        if _ListType.UNORDERED in types or _ListType.ORDERED in types:
            remove_classes.append("list-marker")
        if _ListType.TASK in types:
            remove_classes.append("checkbox")

        line_start = insert.copy()
        line_start.set_line_offset(0)
        markup_start = line_start.copy()

        removed = False
        for cls in remove_classes:
            # Check if have context class at line start
            if cls in self.__buffer.get_context_classes_at_iter(line_start):
                logging.debug("Removing existing marker")
                end = line_start.copy()
                if not iter_forward_to_context_class_removed(end, cls):
                    return False

                # Move past spaces
                first_char = self.__get_char_at_iter(markup_start)
                while first_char.isspace():
                    markup_start.forward_char()
                    first_char = self.__get_char_at_iter(markup_start)

                # Skip this line if we find the marker for a type we're not removing
                if cls == "list-marker":
                    if _ListType.UNORDERED not in types and first_char in ("*", "-", "+"):
                        continue
                    elif _ListType.ORDERED not in types and first_char not in ("*", "-", "+"):
                        continue

                # Delete from the markup start to the end of the context class
                insert_mark = self.__buffer.create_mark(None, markup_start)
                self.__buffer.delete(markup_start, end)
                insert.assign(self.__buffer.get_iter_at_mark(insert_mark))
                self.__buffer.delete_mark(insert_mark)
                removed = True
                break

        return removed

    def __toggle_checkbox_single_line(self, line_start: Gtk.TextIter) -> None:
        # Find checkbox
        check = line_start.copy()
        found = False
        while not check.ends_line():
            check.forward_char()
            if "checkbox-check" in self.__buffer.get_context_classes_at_iter(check):
                found = True
                break

        if not found:
            logging.debug("No checkbox check found to toggle")
            return

        # Toggle
        end = check.copy()
        end.forward_char()
        value = self.__buffer.get_text(check, end, include_hidden_chars=False)
        new_value = "x" if value == " " else " "
        self.__buffer.delete(check, end)
        self.__buffer.insert(check, new_value)

    def __after_blank_line(self, insert: Gtk.TextIter) -> bool:
        if insert.get_line() == 0:
            return False

        line_start = insert.copy()
        line_start.set_line_offset(0)
        location = insert.copy()
        location.backward_line()
        return line_start.get_offset() == location.get_offset() + 1

    def __blank_line_follows(self, insert: Gtk.TextIter) -> bool:
        location = insert.copy()

        # Check for no next line
        location.forward_to_line_end()
        if location.is_end():
            return False

        location.forward_line()
        following_line_offset = location.get_offset()
        location.forward_line()
        # Second match option here caters for end of buffer
        return location.get_offset() in (following_line_offset + 1, following_line_offset)

    def __have_multi_line_selection(self) -> bool:
        if not self.__buffer.get_has_selection():
            return False

        start, end = self.__buffer.get_selection_bounds()
        return start.get_line() != end.get_line()

    def __get_line_quote_marker_depth(self, iter: Gtk.TextIter) -> int:
        if "blockquote" not in self.__buffer.get_context_classes_at_iter(iter):
            return 0

        offset = 0
        loc = iter.copy()
        while not loc.ends_line():
            char = self.__get_line_char_at_iter(loc)
            if char == ">":
                offset += 1
            elif not char.isspace():
                break
            loc.forward_char()
        return offset

    def __get_minimum_quote_marker_depth_across_selection_lines(self) -> int:
        """Get the minimum depth of blockquote (markers) in the selection lines.

        Should only be called with a selection.

        :return: Minimum quote marker depth
        :rtype: int
        """
        if not self.__buffer.get_has_selection():
            logging.warning("No selection, incorrect method use")
            return 0

        offsets = set()
        line_start, end = self.__buffer.get_selection_bounds()
        line_start.set_line_offset(0)
        while line_start.compare(end) < 0 and not line_start.is_end():
            offsets.add(self.__get_line_quote_marker_depth(line_start))
            line_start.forward_line()

        return min(offsets)

    def __add_quote_marker_to_line(self, insert: Gtk.TextIter) -> None:
        self.__move_to_first_quote_marker_on_line(insert)
        self.__buffer.insert(insert, ">")
        self.__insert_space_if_not_exists(insert)

    def __remove_quote_marker_from_line(self, location: Gtk.TextIter) -> None:
        quote_start = location.copy()
        if not self.__move_to_first_quote_marker_on_line(quote_start):
            logging.warning("Couldn't find quote marker")
            return

        # Remove marker
        delete_end = quote_start.copy()
        delete_end.forward_char()
        self.__buffer.delete(quote_start, delete_end)

        # Remove following space if one exists
        if not quote_start.ends_line():
            char = self.__get_line_char_at_iter(quote_start)
            if char.isspace():
                delete_end = quote_start.copy()
                delete_end.forward_char()
                self.__buffer.delete(quote_start, delete_end)

    def __move_to_first_quote_marker_on_line(self, insert: Gtk.TextIter) -> bool:
        found = False
        search = insert.copy()
        while not search.ends_line():
            if self.__get_line_char_at_iter(search) == ">":
                found = True
                break
            search.forward_char()
        if found:
            insert.assign(search)
        return found



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

import os
from typing import Optional
from urllib.parse import quote, urlparse, unquote
from uuid import uuid4

from markdown_it import MarkdownIt
from markdown_it.token import Token
from mdit_py_plugins.attrs import attrs_plugin
from mdit_py_plugins.dollarmath import dollarmath_plugin
from markdown_it_modified_tasklists_plugin.mdit_tasklists_with_elements_plugin import (
    tasklists_with_elements_plugin,
)
from markdown_it_img_figures_plugin.mdit_img_figures_plugin import img_figures_plugin

from iotas.attachment import Attachment
from iotas.note import Note


def parse_to_tokens(
    note: Note,
    exporting: bool,
    tex_support: bool,
) -> tuple[MarkdownIt, list[Token]]:
    """Generate parser tokens for note.

    :param Note note: Note
    :param bool exporting: Whether for export
    :param bool tex_support: Whether to support TeX equations
    :return: Parser and list of parser tokens
    :rtype: tuple[MarkdownIt, list[Token]]
    """
    parser = (
        MarkdownIt("gfm-like")
        .use(tasklists_with_elements_plugin, enabled=not note.read_only)
        .use(attrs_plugin, after=["image"], allowed=["height", "width"])
        .enable("table")
    )
    parser.use(img_figures_plugin, lazyload=not exporting)
    if tex_support:
        parser.use(dollarmath_plugin, renderer=__render_tex)

    if exporting:
        content = get_note_export_content(note, prefix_note_id=False)
    else:
        content = note.content
    return (parser, parser.parse(content))


def get_note_export_content(note: Note, prefix_note_id: bool) -> str:
    """Get note content with image paths updated for export.

    :param Note note: Note
    :param bool prefix_note_id: Whether to prefix the note id
    :return: Updated content
    :rtype: str
    """
    attachments = get_image_attachments_from_note_content(note)
    note_id_prefix = f"{note.id}." if prefix_note_id else ""
    content = note.content

    for attachment in attachments:
        old_path = attachment.path
        new_path = f"attachments/{note_id_prefix}{os.path.basename(old_path)}"
        content = update_image_path_in_markup(content, old_path, new_path)

    return content


def get_image_attachments_from_tokens(note: Note, tokens: list[Token]) -> list[Attachment]:
    """Get list of attachments from parser tokens.

    :param Note note: Note
    :param list[Token] tokens: Parser tokens
    :return: Attachments
    :rtype: list[Attachment]
    """
    images = filter_image_tokens(tokens)
    attachments = []
    for img in images:
        src = img.attrs["src"]
        if not urlparse(src).netloc and not src.strip().startswith("/"):
            path = unquote(src)
            attachment = create_attachment_for_markup_path(note, path)
            attachments.append(attachment)

    return attachments


def get_image_attachments_from_note_content(note: Note) -> list[Attachment]:
    """Get list of attachments from note contents.

    :param Note note: Note
    :return: Attachments
    :rtype: list[Attachment]
    """
    _parser, tokens = parse_to_tokens(note, exporting=False, tex_support=False)
    return get_image_attachments_from_tokens(note, tokens)


def filter_image_tokens(tokens: list[Token]) -> list[Token]:
    """Filter list of parser tokens to images.

    :param list[Token] tokens: Parser tokens
    :return: Image tokens
    :rtype: list[Token]
    """
    images = []

    def add_images_from_children(tokens: list[Token]) -> None:
        for token in tokens:
            if token.children:
                add_images_from_children(token.children)

            if token.type == "image" and "src" in token.attrs:
                images.append(token)

    add_images_from_children(tokens)
    return images


def create_attachment_for_markup_path(note: Note, path: str) -> Attachment:
    """Create an attachment with the provided markup path.

    :param Note note: Note
    :param str path: Attachment markup path, unquoted
    :return: Attachment
    :rtype: Attachment
    """
    attachment = Attachment()
    attachment.note_id = note.id
    attachment.path = path
    attachment.note_remote_id = note.remote_id
    return attachment


def update_image_path_in_markup(content: str, old_path: str, new_path: str) -> str:
    """Update image path in note content.

    :param str content: Note content
    :param str old_path: Previous path, unquoted
    :param str new_path: New path, unquoted
    :return: Updated note content
    :rtype: str
    """
    old_path = quote(old_path)
    new_path = quote(new_path)
    search = f"]({old_path}"

    start_index = content.find(search)
    while start_index >= 0:
        path_start = start_index + 2
        path_end = path_start + len(old_path)

        image_end = __get_image_end_index(content, path_end)

        if image_end is not None:
            # Ensure we haven't matched a link. Clunky but in theory a little less brittle than a
            # weak regex.
            element_start = content.rfind("![", 0, start_index)
            if element_start != -1:
                full_markup = content[element_start:image_end]
                if string_is_image(full_markup):
                    content = content[0:path_start] + new_path + content[path_end:]
                    path_end = path_start + len(new_path)

        start_index = content.find(search, path_end)

    return content


def string_is_image(input: str) -> bool:
    """Whether provided string is markdown for an image.

    :param str input: Markdown string
    :return: Whether an image
    :rtype: bool
    """
    parser = MarkdownIt("gfm-like")
    tokens = parser.parseInline(input)
    if len(tokens) != 1:
        return False
    elif len(tokens[0].children) != 1:
        return False
    elif tokens[0].children[0].type != "image":
        return False
    else:
        return True


def __render_tex(equation: str, options: dict) -> str:
    """Render TeX markdown into HTML element.

    :param str equation: Equation
    :param dict options: Render options
    :return: Generated HTML
    :rtype: str
    """
    span_id = f"tex-{uuid4()}"
    equation = equation.replace("\\", "\\\\")
    js_options = f'{{displayMode: {str(options["display_mode"]).lower()}}}'
    katex_cmd = (
        f"var equation = `{equation}`; "
        + f"katex.render(equation, document.getElementById('{span_id}'), {js_options});"
    )
    return f'<span id="{span_id}"><script>{katex_cmd}</script></span>'


def __get_image_end_index(content: str, path_end: int) -> Optional[int]:
    index = path_end
    found = False
    in_title = False
    while not found:
        if index >= len(content):
            break
        char = content[index]
        if char == "\n":
            break
        elif char == "\r":
            break
        elif char == '"':
            if in_title:
                in_title = False
            else:
                in_title = True
        elif char == ")" and not in_title:
            found = True
        index += 1

    if found:
        return index
    else:
        return None



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

import gi

gi.require_version("WebKit", "6.0")
from gi.repository import Adw, Gdk, GLib, GObject, WebKit, JavaScriptCore

import json
import logging
from threading import Thread
from typing import Callable, Optional
import webbrowser

from markdown_it.token import Token

from iotas.attachment import Attachment
from iotas.attachment_helpers import (
    get_attachment_disk_states,
    get_attachment_filesystem_uri,
    set_attachment_filesystem_uris_on_tokens,
)
from iotas.markdown_helpers import filter_image_tokens
from iotas.config_manager import ConfigManager
from iotas.html_generator import HtmlGenerator
from iotas.markdown_helpers import parse_to_tokens
from iotas.note import Note


class MarkdownRenderView(WebKit.WebView):
    __gtype_name__ = "MarkdownRenderView"

    __gsignals__ = {
        "checkbox-toggled": (GObject.SignalFlags.RUN_FIRST, None, (int, bool)),
        "loaded": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    def __init__(self) -> None:
        super().__init__()
        self.__user_stylesheet = None
        self.__scroll_position = 0.0
        self.__engine_initialised = False
        self.__searching = False
        self.__exporting = False
        self.__loading_attachments: dict[str, Attachment] = {}
        self.__note: Note
        self.__config_manager: Optional[ConfigManager]
        self.__download_attachment: Optional[Callable[[Attachment], bool]] = None
        self.__height_margins_set_for = 0

    def setup(
        self,
        html_generator: HtmlGenerator,
        config_manager: Optional[ConfigManager],
        track_timing: bool,
    ) -> None:
        """Setup for new WebKit instance."""
        self.__html_generator = html_generator
        self.__config_manager = config_manager
        self.__track_timing = track_timing

        self.connect("decide-policy", self.on_decide_policy)
        self.connect("load-changed", self.__on_load_changed)
        self.connect("load-failed", lambda _v, _e, _u, _f: self.__clear_exporting())
        self.connect("context-menu", self.on_context_click)
        content_manager = self.get_user_content_manager()
        content_manager.register_script_message_handler("toPython")
        content_manager.connect("script-message-received", self.__handle_js_message)
        self.__content_manager = content_manager

        if config_manager:
            config_manager.connect_changed(ConfigManager.FONT_SIZE, self.update_style)
            config_manager.connect_changed(
                ConfigManager.MARKDOWN_RENDER_MONOSPACE_FONT_RATIO,
                self.update_style,
            )
            config_manager.connect_changed(ConfigManager.LINE_LENGTH, self.update_style)
            config_manager.connect_changed(
                ConfigManager.EDITOR_HEADER_BAR_VISIBILTY,
                self.update_style,
            )
            config_manager.connect_changed(
                ConfigManager.EDITOR_FORMATTING_BAR_VISIBILTY,
                self.update_style,
            )

        style_manager = Adw.StyleManager.get_default()
        style_manager.connect("notify::high-contrast", lambda _o, _v: self.update_style())

        self.connect("web-process-terminated", self.__log_termination)

        # self.get_settings().set_enable_developer_extras(True)

    def render(
        self,
        note: Note,
        export_format: Optional[str],
        attachment_download_func: Optional[Callable[[Attachment], bool]] = None,
        scroll_position: Optional[float] = None,
    ) -> None:
        """Render view with new markdown.

        :param Note note: Note to render
        :param bool export_format: Export format if exporting, optional
        :param Callable[[[Attachment], bool]] attachment_download_func: Function to
            download missing attachment, optional
        :param float scroll_position: Position to scroll to, optional
        """
        self.update_style()
        if export_format is not None:
            self.exporting = True
        parser, tokens = parse_to_tokens(
            note, exporting=self.exporting, tex_support=self.__get_support_tex()
        )

        attachments_states = get_attachment_disk_states(note, tokens)
        if attachments_states.exists:
            set_attachment_filesystem_uris_on_tokens(tokens, attachments_states.exists)
        if attachments_states.missing and attachment_download_func:
            self.__loading_attachments = attachments_states.missing
            hide_loading_images(tokens, self.__loading_attachments)
        else:
            self.__loading_attachments = {}

        render_func = parser.renderer.render
        content = self.__html_generator.generate(
            note,
            tokens,
            render_func,
            parser.options,
            self.__searching,
            export_format,
            scroll_position,
        )
        self.scroll_position = scroll_position if scroll_position else 0

        self.__parser_tokens = tokens
        self.__download_attachment = attachment_download_func
        self.__load_started_at = GLib.get_monotonic_time()
        self.load_html(content, "file://localhost/")

    def render_retaining_scroll(
        self,
        note: Note,
        export_format: Optional[str],
        attachment_download_func: Optional[Callable[[Attachment], bool]] = None,
    ) -> None:
        """Render view with new markdown retaining scroll position.

        :param Note note: Note to render
        :param bool export_format: Export format if exporting or None
        :param Optional[Callable[[Attachment], bool]] attachment_download_func: Function to
            download missing attachments, optional
        """
        self.render(note, export_format, attachment_download_func, self.scroll_position)

    def update_style(self) -> None:
        """Update the configurable style used in the render."""
        height = self.get_height()
        self.__height_margins_set_for = height
        stylesheet = self.__html_generator.generate_user_stylesheet(self.__searching, height)
        if self.__user_stylesheet is not None:
            self.__content_manager.remove_style_sheet(self.__user_stylesheet)

        self.__user_stylesheet = WebKit.UserStyleSheet(
            stylesheet,
            WebKit.UserContentInjectedFrames.ALL_FRAMES,
            WebKit.UserStyleLevel.USER,
            None,
            None,
        )
        self.__content_manager.add_style_sheet(self.__user_stylesheet)

    def scroll_to_heading(self, line_number: int) -> None:
        """Scroll to heading.

        :param int line_number: Line the heading in source markdown
        """
        script = f"""
        element = document.querySelector('[data-map="{line_number}"]');
        if (element)
            element.scrollIntoView();
        """
        self.evaluate_javascript(
            script, length=-1, world_name=None, source_uri=None, cancellable=None, callback=None
        )

    def show_attached_image(self, attachment: Attachment) -> None:
        """Call Javascript to show downloaded image in render.

        :param Attachment attachment: Attachment
        """
        script = f"""
        elements = document.querySelectorAll('img[src="{attachment.path_quoted}"]');
        for (let i = 0; i < elements.length; i++) {{
             elements[i].src = '{get_attachment_filesystem_uri(attachment)}';
             elements[i].className = '';
        }}
        """
        self.evaluate_javascript(
            script, length=-1, world_name=None, source_uri=None, cancellable=None, callback=None
        )

    def do_size_allocate(self, width: int, height: int, baseline: int) -> None:
        """Allocates widget with a transformation that translates the origin to the position in
        allocation.

        :param int width: Width of the allocation
        :param int height: Height of the allocation
        :param int baseline: The baseline of the child
        """
        WebKit.WebView.do_size_allocate(self, width, height, baseline)
        if abs(height - self.__height_margins_set_for) > 50:
            self.update_style()

    @staticmethod
    def on_decide_policy(
        _web_view: WebKit.WebView,
        decision: WebKit.PolicyDecision,
        decision_type: WebKit.PolicyDecisionType,
    ) -> bool:
        if (
            decision_type == WebKit.PolicyDecisionType.NAVIGATION_ACTION
            and decision.get_navigation_action().is_user_gesture()
        ):
            uri = decision.get_navigation_action().get_request().get_uri()
            webbrowser.open(uri)
            decision.ignore()
            return True
        return False

    @staticmethod
    def on_context_click(
        _web_view: WebKit.WebView,
        menu: WebKit.ContextMenu,
        _event: Gdk.Event,
    ) -> None:
        for item in menu.get_items():
            if item.get_stock_action() in [
                WebKit.ContextMenuAction.DOWNLOAD_LINK_TO_DISK,
                WebKit.ContextMenuAction.GO_BACK,
                WebKit.ContextMenuAction.GO_FORWARD,
                WebKit.ContextMenuAction.OPEN_LINK,
                WebKit.ContextMenuAction.OPEN_LINK_IN_NEW_WINDOW,
                WebKit.ContextMenuAction.RELOAD,
                WebKit.ContextMenuAction.STOP,
            ]:
                menu.remove(item)

    @GObject.Property(type=bool, default=False)
    def exporting(self) -> bool:
        return self.__exporting

    @exporting.setter
    def set_exporting(self, value: bool) -> None:
        self.__exporting = value

    @GObject.Property(type=float)
    def scroll_position(self) -> float:
        return self.__scroll_position

    @scroll_position.setter
    def set_scroll_position(self, value: float) -> None:
        self.__scroll_position = value

    @GObject.Property(type=bool, default=False)
    def searching(self) -> bool:
        return self.__searching

    @searching.setter
    def set_searching(self, value: bool) -> None:
        self.__searching = value
        self.update_style()

    def __on_load_changed(self, _web_view: WebKit.WebView, load_event: WebKit.LoadEvent) -> None:
        if load_event == WebKit.LoadEvent.FINISHED:
            if self.__track_timing:
                duration = GLib.get_monotonic_time() - self.__load_started_at
                if self.__engine_initialised:
                    logging.debug(f"Render took {duration/1000:.2f}ms")
                else:
                    logging.debug(f"WebKit initialisation and render took {duration/1000:.2f}ms")
                self.__engine_initialised = True
            self.emit("loaded")
            if self.__loading_attachments:
                if self.__download_attachment:
                    logging.debug("Start missing images download")
                    thread = Thread(target=self.__download_attachments)
                    thread.daemon = True
                    thread.start()
                else:
                    logging.warning(f"Missing {len(self.__loading_attachments)} attachments")

    def __download_attachments(self) -> None:
        if not self.__download_attachment:
            return
        for attachment in self.__loading_attachments.values():
            self.__download_attachment(attachment)

    def __clear_exporting(self) -> None:
        self.exporting = False

    def __handle_js_message(
        self, _manager: WebKit.UserContentManager, result: JavaScriptCore.Value
    ) -> None:
        js = result.to_json(2)
        message = json.loads(js)
        if message["type"] == "checkbox":
            self.__handle_checkbox_toggle(message)
        elif message["type"] == "scrollPosition":
            if "position" not in message or message["position"] is None:
                logging.warning("Received scroll position message with no position")
                return
            self.scroll_position = message["position"]

    def __handle_checkbox_toggle(self, message: dict) -> None:
        search_id = message["id"]
        new_value = message["checked"]

        def match(token: Token, search_id: str) -> bool:
            search_str = f'id="{search_id}"'
            return token.type == "html_inline" and search_str in token.content

        def handle_match(line: int, new_value: bool) -> None:
            self.emit("checkbox-toggled", line, new_value)

        line: int | None = None
        for token in self.__parser_tokens:
            if token.map is not None:
                line = token.map[0]
            if match(token, search_id):
                logging.debug(
                    f'Setting checkbox on line {line} with id "{search_id}" to {new_value}'
                )
                assert line is not None
                handle_match(line, new_value)
                return
            if token.children is not None:
                for child in token.children:
                    if match(child, search_id):
                        if token.map is not None:
                            line = token.map[0]
                        logging.debug(
                            f'Setting checkbox on line {line} with id "{search_id}" to {new_value}'
                        )
                        assert line is not None
                        handle_match(line, new_value)
                        return

    def __log_termination(
        self, _web_view: WebKit.WebView, reason: WebKit.WebProcessTerminationReason
    ) -> None:
        if self.__track_timing:
            logging.debug(f"WebKit process terminated for {reason}")
        self.__engine_initialised = False

    def __get_support_tex(self) -> bool:
        if not self.__config_manager:
            return False
        else:
            return self.__config_manager.markdown_tex_support


def hide_loading_images(tokens: list[Token], transferring: dict[str, Attachment]) -> None:
    """Hide server transferring images in parser tokens.

    :param list[Token] tokens: Parser tokens
    :param dict[str, Attachment] transferring: Transferring attachments
    """
    images = filter_image_tokens(tokens)
    for img in images:
        src = img.attrs["src"]
        if src in transferring:
            img.attrs["class"] = "transferring"



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from abc import ABC, abstractmethod

from typing import Callable, Optional

from markdown_it.token import Token
from markdown_it.utils import EnvType, OptionsDict

from iotas.note import Note


class HtmlGenerator(ABC):
    """HTML generator base."""

    RESOURCE_CSS_PATH = "/media/css/web/markdown.css"
    RESOURCE_KATEX_CSS_PATH = "/media/css/web/katex.min.css"
    RESOURCE_KATEX_JS_PATH = "/media/js/katex.min.js"

    @abstractmethod
    def generate(
        self,
        note: Note,
        tokens: list[Token],
        render_func: Callable[[list[Token], OptionsDict, EnvType], str],
        parser_options: OptionsDict,
        searching: bool,
        export_format: Optional[str],
        scroll_position: Optional[float] = None,
    ) -> str:
        """Generator HTML for note.

        :param Note note: Note to render
        :param list[Token] tokens: Parser tokens
        :param Callable[[list[Token], OptionsDict, EnvType], str] render_func: Render function
        :param OptionsDict parser_options: Parser options
        :param bool searching: Whether search CSS should be included
        :param Optional[str] export_format: Export format, if using
        :param Optional[float] scroll_position: Position to scroll to
        :return: Generated HTML
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def generate_user_stylesheet(self, searching: bool, height: int) -> str:
        """Generate part of stylesheet based on state (configuration etc).

        :param bool searching: Whether searching
        :param int height: View height
        :return: stylesheet
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def update_font_family(self, family: str) -> None:
        """Update the font family.

        :param str family: New font family
        """
        raise NotImplementedError()



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from gi.repository import GObject

import logging
import re
import time
from typing import Optional, Self

from iotas.text_utils import sanitise_path


class DirtyFields:
    title = False
    content = False
    excerpt = False
    favourite = False
    category = False
    etag = False
    remote_id = False
    last_modified = False
    locally_deleted = False
    read_only = False

    def empty(self) -> bool:
        """Whether no fields have been modified.

        :return: Whether there are modifications
        :rtype: bool
        """
        return (
            not self.title
            and not self.content
            and not self.excerpt
            and not self.favourite
            and not self.category
            and not self.etag
            and not self.remote_id
            and not self.last_modified
            and not self.locally_deleted
            and not self.read_only
        )


class _RecentETagList:
    """
    List of recent known ETags.
    """

    SIZE = 50

    def __init__(self) -> None:
        self.__tags: list[str] = []

    def add(self, tag: str) -> None:
        """Add an ETag.

        :param str tag: ETag
        """
        if not tag.strip():
            return
        self.__trim()
        self.__tags.append(tag)

    def contains(self, tag: str) -> bool:
        """Check if ETag is in list.

        :param str tag: ETag
        :return: Whether in list
        :rtype: bool
        """
        return tag in self.__tags

    def __trim(self) -> None:
        while len(self.__tags) >= self.SIZE:
            del self.__tags[0]


class Note(GObject.Object):
    __gsignals__ = {
        "remote-content-update": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    EXCERPT_LENGTH = 200
    EXCERPT_LINE_SEPARATOR = "   "

    def __init__(self, new_note: bool = False) -> None:
        super().__init__()
        self.__title = ""
        self.__excerpt = ""
        self.__content: Optional[str] = None
        self.__category = ""
        self.__favourite = False
        self.__id = -1
        self.__last_modified = 0
        self.__read_only = False
        self.__dirty = False
        self.__dirty_while_saving = False
        self.__server_sanitised_title: Optional[str] = None
        self.__etag = ""
        self.__remote_id = -1
        self.__title_is_top_line = new_note
        self.__saving = False
        self.__locally_deleted = False
        self.__handling_conflict = False
        self.__recent_tags = _RecentETagList()

    def duplicate(self) -> Self:
        """Duplicate note.

        :return: Duplicate note
        :rtype: Self
        """
        n = Note()
        n.title = self.title
        n.excerpt = self.excerpt
        n.content = self.content
        n.favourite = self.favourite
        n.category = self.category
        return n

    def flag_changed(self, update_last_modified: bool = True) -> None:
        """Flag that the note has changed.

        dirty, last_modified and dirty_while_saving will be updated as applicable.
        """
        if not self.dirty:
            self.dirty = True
        if update_last_modified:
            self.last_modified = int(time.time())
        if self.saving:
            self.dirty_while_saving = True

    def regenerate_excerpt(self) -> bool:
        """Regenerates the excerpt from the content.

        :return: Whether the excerpt has changed
        :rtype: bool
        """
        if self.__content is None:
            logging.warning(
                "Can't generate excerpt without content populated ({})".format(self.__title)
            )
            return False
        new_excerpt = self.__content.strip()
        if new_excerpt != "":
            # Remove any heading markdown before first line
            new_excerpt = re.sub(r"^#* ", "", new_excerpt).strip()

            title = self.__title.strip()
            if len(title) and new_excerpt.startswith(title):
                # Remove title from start of content
                new_excerpt = new_excerpt[len(title) :]

            new_excerpt = re.sub(r"^\s*[-*+] \[ \]", "☐", new_excerpt, flags=re.MULTILINE)
            new_excerpt = re.sub(r"^\s*[-*+] \[x\]", "☑", new_excerpt, flags=re.MULTILINE)
            new_excerpt = re.sub(r"^\s*[-*+] (?![-*+[])", "• ", new_excerpt, flags=re.MULTILINE)

            new_excerpt = new_excerpt.replace("\n", self.EXCERPT_LINE_SEPARATOR).strip()
            if len(new_excerpt) > self.EXCERPT_LENGTH:
                new_excerpt = new_excerpt[: self.EXCERPT_LENGTH]

        changed = new_excerpt != self.excerpt
        if changed:
            self.excerpt = new_excerpt
        return changed

    def repopulate_meta_from_sync_update(self, update_dict: dict) -> None:
        """Repopulate note metadata from update over REST API JSON.

        :param dict update_dict: Nextcloud Notes API JSON
        """
        self.__last_modified = update_dict["modified"]
        self.etag = update_dict["etag"]
        self.__remote_id = update_dict["id"]

        # Handle title and category updates from sanitisation on server side
        if self.title != update_dict["title"]:
            if self.title_is_top_line:
                self.server_sanitised_title = update_dict["title"]
            else:
                self.title = update_dict["title"]
        if self.category != update_dict["category"]:
            self.category = update_dict["category"]

    def update_from_remote_sync(self, note_dict: dict) -> DirtyFields:
        """Update from remote sync JSON dict and flag which fields are updated.

        :param dict note_dict: Nextcloud Notes API JSON
        :return: Which fields were updated
        :rtype: DirtyFields
        """
        f = DirtyFields()
        if self.__last_modified != note_dict["modified"]:
            self.last_modified = note_dict["modified"]
            f.last_modified = True
        if self.__etag != note_dict["etag"]:
            self.etag = note_dict["etag"]
            f.etag = True
        if self.__read_only != note_dict["readonly"]:
            self.read_only = note_dict["readonly"]
            f.read_only = True
        if self.__favourite != note_dict["favorite"]:
            self.favourite = note_dict["favorite"]
            f.favourite = True
        if self.__category != note_dict["category"]:
            self.category = note_dict["category"]
            f.category = True
        if self.__title != note_dict["title"]:
            self.title = note_dict["title"]
            f.title = True
        if "ignore-content" not in note_dict and self.__content != note_dict["content"]:
            self.content = note_dict["content"]
            f.content = True
            # Always called on UI thread
            self.emit("remote-content-update")
        if (f.title or f.content) and self.content:
            self.regenerate_excerpt()
            f.excerpt = True
        return f

    def update_title_from_top_line(self) -> None:
        """Update the title from the top line of the content."""
        if not self.__title_is_top_line:
            return
        if self.__content is not None and self.__content.strip() != "":
            lines = self.__content.strip().split("\n")
            title = lines[0].strip()
            title = title.strip("#")
            title = title.strip()
            title = sanitise_path(title)
            if len(title) > 200:
                title = title[:200]
            self.title = title

    def identical_excepting_sync_meta(self, note2: Note) -> bool:
        """Check whether notes are identical, apart from sync meta fields.

        :param Note note2: Note to compare to
        :return: Whether matching
        :rtype: bool
        """
        return self.identical_non_sync_meta(note2) and self.content == note2.content

    def identical_to(self, note2: Note) -> bool:
        """Check whether notes are identical.

        :param Note note2: Note to compare to
        :return: Whether identical
        :rtype: bool
        """
        return (
            self.identical_excepting_sync_meta(note2)
            and self.last_modified == note2.last_modified
            and self.dirty == note2.dirty
            and self.etag == note2.etag
        )

    def identical_non_sync_meta(self, note2: Note) -> bool:
        """Check whether notes have identical meta fields, excluding sync meta.

        :param Note note2: Note to compare to
        :return: Whether matching
        :rtype: bool
        """
        return (
            self.title == note2.title
            and self.favourite == note2.favourite
            and self.category == note2.category
            and self.locally_deleted == note2.locally_deleted
            and self.remote_id == note2.remote_id
        )

    def etag_is_recent(self, etag: str) -> bool:
        """Check whether etag matches a known recent push.

        :param str etag: ETag to compare
        :return: Whether known
        :rtype: bool
        """
        return self.__recent_tags.contains(etag)

    def to_nextcloud_dict(self) -> dict:
        """Create an API dict object for the note, for debugging.

        :return: Note as API results dict
        :rtype: dict
        """
        return {
            "id": self.remote_id,
            "title": self.title,
            "content": self.content,
            "category": self.category,
            "favorite": self.favourite,
            "modified": self.last_modified,
            "etag": self.etag,
            "readonly": self.read_only,
        }

    @staticmethod
    def from_backup(meta: dict, content: str, include_sync_attributes: bool) -> Optional[Note]:
        """Create note from backup dict.

        :param dict meta: Backup metadata
        :param str content: Content
        :param bool include_sync_attributes: Whether to include sync attributes
        :return: The note or None on failure
        :rtype: Optional[Self]
        """
        note = Note()
        try:
            note.title = meta["Title"]
            note.content = content
            note.regenerate_excerpt()
            note.category = meta["Category"]
            note.favourite = meta["Favourite"]
            note.last_modified = meta["LastModified"]
            note.locally_deleted = meta["LocallyDeleted"]
            note.dirty = meta["Dirty"]
            if include_sync_attributes:
                note.etag = meta["ETag"]
                note.remote_id = meta["RemoteId"]
        except KeyError:
            return None
        return note

    @staticmethod
    def from_nextcloud(note_dict: dict) -> Note:
        """Create note from dict originating from REST API JSON.

        :param dict note_dict: Nextcloud Notes API JSON
        :return: The note
        :rtype: Self
        """
        note = Note()
        note.title = note_dict["title"]
        note.content = note_dict["content"]
        note.regenerate_excerpt()
        note.category = note_dict["category"]
        note.favourite = note_dict["favorite"]
        note.last_modified = note_dict["modified"]
        note.etag = note_dict["etag"]
        note.remote_id = note_dict["id"]
        note.read_only = note_dict["readonly"]
        return note

    @GObject.Property(type=str, default="")
    def title(self) -> str:
        return self.__title

    @title.setter
    def set_title(self, value: str) -> None:
        self.__title = value
        self.__server_sanitised_title = None

    @GObject.Property(type=str, default="")
    def excerpt(self) -> str:
        return self.__excerpt

    @excerpt.setter
    def set_excerpt(self, value: str) -> None:
        self.__excerpt = value

    @GObject.Property(type=str)
    def content(self) -> Optional[str]:
        return self.__content

    @content.setter
    def set_content(self, value: str) -> None:
        self.__content = value

    @GObject.Property(type=bool, default=False)
    def content_loaded(self) -> bool:
        """Whether the note's contents are populated.

        :return: Whether populated
        :rtype: bool
        """
        return self.__content is not None

    @GObject.Property(type=bool, default=False)
    def favourite(self) -> bool:
        return self.__favourite

    @favourite.setter
    def set_favourite(self, value: bool) -> None:
        self.__favourite = value

    @GObject.Property(type=str)
    def category(self) -> str:
        return self.__category

    @category.setter
    def set_category(self, value: str) -> None:
        self.__category = value

    @GObject.Property(type=bool, default=False)
    def has_id(self) -> bool:
        """Whether the note has an id (as a result of being persisted to the database).

        :return: Whether the note has been persisted to the database
        :rtype: bool
        """
        return self.__id != -1

    @GObject.Property(type=int, default=-1)
    def id(self) -> int:
        return self.__id

    @id.setter
    def set_id(self, value: int) -> None:
        self.__id = value

    @GObject.Property(type=int, default=0)
    def last_modified(self) -> int:
        return self.__last_modified

    @last_modified.setter
    def set_last_modified(self, value: int) -> None:
        self.__last_modified = value

    @GObject.Property(type=bool, default=False)
    def read_only(self) -> bool:
        return self.__read_only

    @read_only.setter
    def set_read_only(self, value: bool) -> None:
        self.__read_only = value

    @GObject.Property(type=bool, default=False)
    def dirty(self) -> bool:
        return self.__dirty

    @dirty.setter
    def set_dirty(self, value: bool) -> None:
        self.__dirty = value

    @GObject.Property(type=bool, default=False)
    def dirty_while_saving(self) -> bool:
        return self.__dirty_while_saving

    @dirty_while_saving.setter
    def set_dirty_while_saving(self, value: bool) -> None:
        self.__dirty_while_saving = value

    @GObject.Property(type=bool, default=False)
    def saving(self) -> bool:
        return self.__saving

    @saving.setter
    def set_saving(self, value: bool) -> None:
        self.__saving = value

    @GObject.Property(type=bool, default=False)
    def locally_deleted(self) -> bool:
        return self.__locally_deleted

    @locally_deleted.setter
    def set_locally_deleted(self, value: bool) -> None:
        self.__locally_deleted = value

    @GObject.Property(type=bool, default=False)
    def handling_conflict(self) -> bool:
        return self.__handling_conflict

    @handling_conflict.setter
    def set_handling_conflict(self, value: bool) -> None:
        self.__handling_conflict = value

    @GObject.Property(type=str, default="")
    def server_sanitised_title(self) -> Optional[str]:
        return self.__server_sanitised_title

    @server_sanitised_title.setter
    def set_server_sanitised_title(self, value: str) -> None:
        self.__server_sanitised_title = value

    @GObject.Property(type=str, default="")
    def etag(self) -> str:
        return self.__etag

    @etag.setter
    def set_etag(self, value: str) -> None:
        self.__etag = value
        self.__recent_tags.add(value)

    @GObject.Property(type=int, default=-1)
    def remote_id(self) -> int:
        return self.__remote_id

    @remote_id.setter
    def set_remote_id(self, value: int) -> None:
        self.__remote_id = value

    @GObject.Property(type=bool, default=False)
    def has_remote_id(self) -> bool:
        """Whether the note has a remote id (as a result of being synced to a remote server).

        :return: Whether the note has, at some point, been synced to a remote server
        :rtype: bool
        """
        return self.__remote_id >= 0

    @GObject.Property(type=bool, default=False)
    def title_is_top_line(self) -> bool:
        return self.__title_is_top_line

    @title_is_top_line.setter
    def set_title_is_top_line(self, value: bool) -> None:
        self.__title_is_top_line = value

    @GObject.Property(type=bool, default=False)
    def new_and_empty(self) -> bool:
        """Check if the note is new and has no content or title.

        :return: Whether new and empty
        :rtype: bool
        """
        return (
            self.id == -1
            and self.title.strip() == ""
            and self.content is not None
            and self.content.strip() == ""
        )



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Gtk, GtkSource

import re
from typing import Callable, Optional

from iotas.ordered_list_utils import (
    check_line_for_ordered_list_item,
    calculate_ordered_list_index,
    format_ordered_list_item,
)

# k is the token, v is the continuation
_BULLET_LIST_TOKENS = {
    "- [ ] ": "- [ ] ",
    "- [x] ": "- [ ] ",
    "+ [ ] ": "+ [ ] ",
    "+ [x] ": "+ [ ] ",
    "* [ ] ": "* [ ] ",
    "* [x] ": "* [ ] ",
    "- ": "- ",
    "+ ": "+ ",
    "* ": "* ",
}


def check_and_extend_list(buffer: GtkSource.Buffer) -> bool:
    """Extend any list after a newline is inserted.

    :param GtkSource.Buffer buffer: Buffer to work on
    """
    handled = False
    if buffer.get_has_selection():
        insert_iter, _unused = buffer.get_selection_bounds()
    else:
        mark = buffer.get_insert()
        insert_iter = buffer.get_iter_at_mark(mark)

    line_start = insert_iter.copy()
    line_start.set_line_offset(0)
    previous_line = line_start.get_text(insert_iter)

    def check_and_delete_selection(insert_iter, line_start) -> tuple[Gtk.TextIter, Gtk.TextIter]:
        if buffer.get_has_selection():
            buffer.delete_selection()
            mark = buffer.get_insert()
            insert_iter = buffer.get_iter_at_mark(mark)
            line_start = insert_iter.copy()
            line_start.set_line_offset(0)
        return insert_iter, line_start

    # TODO look at changing to use the GtkSourceLanguage context?
    regex_match = _check_line_for_bullet_list_item(previous_line)
    if regex_match is not None:
        insert_iter, line_start = check_and_delete_selection(insert_iter, line_start)
        handled = _extend_bullet_list(buffer, regex_match.group(), insert_iter, line_start)
    else:
        regex_match = check_line_for_ordered_list_item(previous_line)
        if regex_match is not None:
            insert_iter, line_start = check_and_delete_selection(insert_iter, line_start)
            handled = _extend_ordered_list(buffer, regex_match.groups(), insert_iter, line_start)

    return handled


def increase_indentation(buffer: GtkSource.Buffer) -> None:
    """Increase indentation on a buffer line or selection.

    :param Gtk.Buffer buffer: Buffer to work on
    """
    _handle_tab(buffer, True)


def decrease_indentation(buffer: GtkSource.Buffer) -> None:
    """Decrease indentation on a buffer line or selection.

    :param Gtk.Buffer buffer: Buffer to work on
    """
    _handle_tab(buffer, False)


def _handle_tab(buffer: GtkSource.Buffer, increase: bool) -> None:
    buffer.begin_user_action()
    if buffer.get_has_selection():
        begin, end = buffer.get_selection_bounds()
        begin.order(end)
        multi_line = "\n" in buffer.get_text(begin, end, False)
        if multi_line:
            line_iter = begin.copy()
            line_mark = buffer.create_mark(None, line_iter)
            end_mark = buffer.create_mark(None, end)
            while line_iter.compare(end) < 0:
                _modify_single_line_indent(buffer, line_iter, increase, multi_line)
                line_iter = buffer.get_iter_at_mark(line_mark)
                line_iter.forward_line()
                buffer.delete_mark(line_mark)
                line_mark = buffer.create_mark(None, line_iter)
                end = buffer.get_iter_at_mark(end_mark)
            buffer.delete_mark(line_mark)
            buffer.delete_mark(end_mark)
        else:
            _modify_single_line_indent(buffer, begin, increase, multi_line)
    else:
        mark = buffer.get_insert()
        begin = buffer.get_iter_at_mark(mark)
        _modify_single_line_indent(buffer, begin, increase, False)
    buffer.end_user_action()


def _extend_bullet_list(
    buffer: GtkSource.Buffer,
    matched_list_item: str,
    insert_iter: Gtk.TextIter,
    line_start: Gtk.TextIter,
) -> bool:
    previous_line = line_start.get_text(insert_iter)

    # If the list already continues after our newline, don't add the next item markup.
    # Caters for accidentally deleting a line break in the middle of a list and then
    # pressing enter to revert that.
    cur_line_end = insert_iter.copy()
    if not cur_line_end.ends_line():
        cur_line_end.forward_to_line_end()
    cur_line_text = insert_iter.get_text(cur_line_end)
    if _check_line_for_bullet_list_item(cur_line_text) is not None:
        return False

    def generate_sequence_previous_item() -> Optional[str]:
        return matched_list_item

    buffer.begin_user_action()
    empty_list_line = _inserted_empty_item_at_end_of_list(
        previous_line, matched_list_item, line_start, generate_sequence_previous_item
    )
    if empty_list_line:
        # An empty list line has been entered, remove it from the list end
        buffer.delete(line_start, insert_iter)
        buffer.insert_at_cursor("\n")
    else:
        dict_key = matched_list_item.lstrip()
        spacing = matched_list_item[0 : len(matched_list_item) - len(dict_key)]
        new_entry = f"\n{spacing}{_BULLET_LIST_TOKENS[dict_key]}"
        buffer.insert_at_cursor(new_entry)
    buffer.end_user_action()
    return True


def _extend_ordered_list(
    buffer: GtkSource.Buffer,
    regex_groups: tuple,
    insert_iter: Gtk.TextIter,
    line_start: Gtk.TextIter,
) -> bool:
    spacing = regex_groups[0]
    index = regex_groups[1]
    marker = regex_groups[2]
    previous_line = line_start.get_text(insert_iter)

    def generate_sequence_previous_item() -> Optional[str]:
        # Verify there's more than one list item. This allows inserting lines with list start
        # tokens and nothing else.
        previous_index = calculate_ordered_list_index(index, -1)
        if previous_index is None:
            return None
        else:
            return format_ordered_list_item(spacing, previous_index, marker)

    buffer.begin_user_action()
    empty_list_line = _inserted_empty_item_at_end_of_list(
        previous_line, "".join(regex_groups), line_start, generate_sequence_previous_item
    )
    if empty_list_line:
        # An empty list line has been entered, remove it from the list end
        buffer.delete(line_start, insert_iter)
        buffer.insert_at_cursor("\n")
        buffer.end_user_action()
    else:
        sequence_next = calculate_ordered_list_index(index, 1)
        # Handle value of Z, ending sequence
        if sequence_next is None:
            buffer.end_user_action()
            return False
        new_entry = "\n" + format_ordered_list_item(spacing, sequence_next, marker)
        buffer.insert(insert_iter, new_entry)

        _increment_any_following_ordered_list_items(
            buffer, insert_iter, sequence_next, marker, spacing
        )

        buffer.end_user_action()

    return True


def _increment_any_following_ordered_list_items(
    buffer: GtkSource.Buffer, insert_iter: Gtk.TextIter, index: str, marker: str, spacing: str
) -> None:
    cursor_mark = buffer.create_mark(None, insert_iter)

    current_entry = format_ordered_list_item(spacing, index, marker)

    while True:
        # Check if the next line continues the list
        if not insert_iter.forward_line():
            break
        line_end = insert_iter.copy()
        line_end.forward_to_line_end()
        text = buffer.get_text(insert_iter, line_end, include_hidden_chars=False)
        if not text.startswith(current_entry):
            break

        next_index = calculate_ordered_list_index(index, 1)
        # Handle value of Z, ending sequence
        if not next_index:
            break
        index = next_index
        entry_next = format_ordered_list_item(spacing, index, marker)

        # Delete existing
        end = insert_iter.copy()
        end.forward_chars(len(current_entry))
        buffer.delete(insert_iter, end)

        # Insert new
        buffer.insert(insert_iter, entry_next)

        current_entry = entry_next

    buffer.place_cursor(buffer.get_iter_at_mark(cursor_mark))
    buffer.delete_mark(cursor_mark)


def _modify_single_line_indent(
    buffer: GtkSource.Buffer, location: Gtk.TextIter, increase: bool, multi_line: bool
) -> None:
    line_start = location.copy()
    line_start.set_line_offset(0)
    line_end = location.copy()
    if not line_end.ends_line():
        line_end.forward_to_line_end()
    line_contents = buffer.get_text(line_start, line_end, False)

    # Don't process empty lines
    if multi_line and line_contents == "":
        return

    list_item = (
        _check_line_for_bullet_list_item(line_contents) is not None
        or check_line_for_ordered_list_item(line_contents) is not None
    )
    indent_chars = "  " if list_item else "\t"

    if increase:
        if multi_line or list_item:
            buffer.insert(line_start, indent_chars)
        else:
            buffer.insert(location, indent_chars)
    else:
        if line_contents.startswith(indent_chars):
            end_deletion = line_start.copy()
            end_deletion.forward_chars(len(indent_chars))
            buffer.delete(line_start, end_deletion)


def _inserted_empty_item_at_end_of_list(
    line_text: str,
    matched_list_item: str,
    previous_line_start: Gtk.TextIter,
    fetch_sequence_previous: Callable[[], Optional[str]],
) -> bool:
    """Determine whether we have inserted an empty item at the end of a list

    The list needs to have at least one previous item.

    :param str line_text: The full line text
    :param str matched_list_item: Our matched list marker
    :param Gtk.TextIter previous_line_start: The start of the previous line
    :param Callable[], Optional[str]] fetch_sequence_previous: A function to fetch a previous
        line's marker
    :return: Whether we've inserted an empty line
    :rtype: bool
    """

    # Check if entered line contains only the list item markup
    if line_text.strip() != matched_list_item.strip():
        return False

    # Fetch what a previous item in the sequence would have been, for the check below. The
    # callable is used to prevent calculating this before we even know if match above will
    # pass (while sharing logic for bullet and ordered lists).
    sequence_previous_item = fetch_sequence_previous()
    if sequence_previous_item is None:
        return False

    # Verify there's more than one list item. This allows inserting lines with list start
    # tokens and nothing else.
    two_prev_start = previous_line_start.copy()
    two_prev_start.backward_line()
    two_prev_start.set_line_offset(0)
    two_prev_line = two_prev_start.get_text(previous_line_start)
    return two_prev_line.startswith(sequence_previous_item)


def _check_line_for_bullet_list_item(linetext: str) -> Optional[re.Match[str]]:
    # Avoid misidentifying our horizontal rule markup as an ordered list line. This should be made
    # to handle other valid horizontal rule forms, and be integrated into the regex below.
    if linetext == "- - -":
        return None

    term = r"^\s*("
    escaped_tokens = []
    for token in _BULLET_LIST_TOKENS.keys():
        escaped_tokens.append(re.escape(token))
    term += "|".join(escaped_tokens)
    term += ")"
    return re.search(term, linetext)



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import GObject, Gio, Graphene, Gtk

from typing import Any, Callable, Generator, NamedTuple, Optional

from iotas.category import Category
from iotas.config_manager import ConfigManager
from iotas.index_row import IndexRow
from iotas.note import Note
from iotas.note_list_model import NoteListModelTimeFiltered, NoteListModelSearch
from iotas.note_manager import NoteManager


class SectionInfo(NamedTuple):
    model: Gio.ListModel
    listbox: Gtk.ListBox
    section: Gtk.Box


class FocusedRow(NamedTuple):
    listbox: Gtk.ListBox
    indexx: int


@Gtk.Template(resource_path="/org/gnome/World/Iotas/ui/index_note_list.ui")
class IndexNoteList(Gtk.Box):
    __gtype_name__ = "IndexNoteList"
    __gsignals__ = {
        "note-opened": (GObject.SignalFlags.RUN_FIRST, None, (Note,)),
        "note-checkbox-activated": (GObject.SignalFlags.RUN_FIRST, None, (Note,)),
        "note-checkbox-deactivated": (GObject.SignalFlags.RUN_FIRST, None, (Note,)),
        "search-model-emptied": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    MINIMUM_SEARCH_PAGE_NOTE_COUNT = 15
    ROW_HEIGHT = 80

    _favourites_listbox: Gtk.ListBox = Gtk.Template.Child()
    _favourites_section: Gtk.Box = Gtk.Template.Child()
    _today_listbox: Gtk.ListBox = Gtk.Template.Child()
    _today_section: Gtk.Box = Gtk.Template.Child()
    _yesterday_listbox: Gtk.ListBox = Gtk.Template.Child()
    _yesterday_section: Gtk.Box = Gtk.Template.Child()
    _week_listbox: Gtk.ListBox = Gtk.Template.Child()
    _week_section: Gtk.Box = Gtk.Template.Child()
    _month_listbox: Gtk.ListBox = Gtk.Template.Child()
    _month_section: Gtk.Box = Gtk.Template.Child()
    _last_month_listbox: Gtk.ListBox = Gtk.Template.Child()
    _last_month_section: Gtk.Box = Gtk.Template.Child()
    _older_notes_loader: Gtk.Box = Gtk.Template.Child()
    _older_notes_listbox: Gtk.ListBox = Gtk.Template.Child()
    _older_notes_section: Gtk.Box = Gtk.Template.Child()
    _show_more_button: Gtk.Box = Gtk.Template.Child()
    _search_listbox: Gtk.ListBox = Gtk.Template.Child()
    _search_section: Gtk.Box = Gtk.Template.Child()

    def __init__(self) -> None:
        super().__init__()

        self.__config_manager = ConfigManager.get_default()

        self.__selecting = False
        self.__searching = False
        self.__search_full_set: list[int] = []
        self.__older_notes_displayed = False
        self.__search_base_note_count = self.MINIMUM_SEARCH_PAGE_NOTE_COUNT
        self.__main_sections: list[SectionInfo] = []
        self.__extra_listboxes: list[Gtk.Listbox] = []

    def setup(
        self,
        note_manager: NoteManager,
        listbox_keypress_func: Callable,
    ) -> None:
        """Perform initial setup.

        :param NoteManager note_manager: The note manager
        :param Callable listbox_keypress_func: Function to call on listbox keypress
        """
        self.__listbox_keypress_func = listbox_keypress_func

        self.__populate_sections_list(note_manager)
        self.__older_notes_model: Optional[NoteListModelTimeFiltered] = None
        self.__search_model: Optional[NoteListModelSearch] = None

        for section in self.__main_sections:
            section.listbox.bind_model(section.model, self.__create_row, None)
            section.model.connect("items-changed", self.__on_model_items_changed, section.section)
            self.__add_listbox_keypress_controller(section.listbox)

    def update_search_pagesize_for_height(self, height: int) -> None:
        """Recalculate number of results to show on first search "page".

        :param int height: The new window height
        """
        self.__search_base_note_count = max(
            self.MINIMUM_SEARCH_PAGE_NOTE_COUNT, int((height / self.ROW_HEIGHT) + 5)
        )

    def clear_selections(self) -> None:
        """Clear all selections in the list."""
        self.__clear_selections_excluding(None)

    def populate_older_notes(self, model: NoteListModelTimeFiltered) -> bool:
        """Populate and display notes older than two months.

        Populates the remaining notes which haven't been shown in the index as they are beyond the
        two month old mark.

        :param NoteListModelTimeFiltered model: Model for the older notes
        :return: Whether the model contains any notes after population
        :rtype: bool
        """
        if self.__older_notes_model is None:
            self.__older_notes_model = model
            self.__extra_listboxes.append(self._older_notes_listbox)
            self.__add_listbox_keypress_controller(self._older_notes_listbox)

        return self.__display_older_notes()

    def populate_search(self, model: NoteListModelSearch) -> None:
        """Populate search model, binding to list box etc.

        :param NoteListModelSearch model: Search model
        """
        if not self.__search_model:
            self.__search_model = model
            self._search_listbox.bind_model(model, self.__create_row, None)
            self.__extra_listboxes.append(self._search_listbox)
            self.__add_listbox_keypress_controller(self._search_listbox)
            model.connect("items-changed", self.__on_search_items_changed)

    def refresh_section_visibility(self, category: Category) -> None:
        """Refresh visibility of list sections.

        :param Category category: The current category
        """
        if self.searching:
            for section in self.__main_sections:
                section.section.set_visible(False)
            self._older_notes_section.set_visible(False)
            self._older_notes_loader.set_visible(False)
            self._search_section.set_visible(True)
        else:
            for section in self.__main_sections:
                section.section.set_visible(len(section.model) > 0)
            self._older_notes_section.set_visible(
                self.older_notes_displayed
                and self.__older_notes_model is not None
                and len(self.__older_notes_model) > 0
            )
            self._older_notes_loader.set_visible(not self.older_notes_displayed)
            self._search_section.set_visible(False)

    def get_selected_note(self) -> Optional[Note]:
        """Fetch the currently selected note.

        :return: The selected note, if applicable
        :rtype: Optional[Note]
        """
        selected = None
        for listbox in self.__visible_listboxes():
            row = listbox.get_selected_row()
            if row:
                selected = row.get_child().note
                break
        return selected

    def move_focus_to_list_top(self) -> None:
        """Move focus to the top of the first visible section."""
        for listbox in self.__visible_listboxes():
            row = listbox.get_row_at_index(0)
            if row is None:
                return
            row.grab_focus()
            listbox.select_row(row)
            break

    def move_focus_to_first_visible(self, below: float) -> bool:
        """Move focus to first row visible below threshold.

        :param float below: Minimum vertical position
        """
        done = False
        for listbox in self.__visible_listboxes():
            i = 0
            row = listbox.get_row_at_index(i)
            while row is not None and not done:
                common, pt = row.compute_point(self, Graphene.point_zero())
                if common and pt.y >= below:
                    row.grab_focus()
                    listbox.select_row(row)
                    done = True
                else:
                    i += 1
                    row = listbox.get_row_at_index(i)
            if done:
                break

        return done

    def refocus_selected_row(self) -> bool:
        """Refocus the selected row.

        :return: Whether a selected row was found
        :rtype: bool
        """
        focused = False
        for listbox in self.__visible_listboxes():
            row = listbox.get_selected_row()
            if row:
                row.grab_focus()
                focused = True
                break
        return focused

    def get_current_selection(self) -> Optional[FocusedRow]:
        """Get currently focused row.

        :return: Selected listbox and row index, optional
        :rtype: FocusedRow
        """
        for listbox in self.__visible_listboxes():
            row = listbox.get_selected_row()
            if row:
                return FocusedRow(listbox, row.get_index())

        return None

    def place_focus_after_change(self, focused_row: FocusedRow) -> None:
        """Refocus row or sibling.

        Attempts in order:
        1. Any row in the listbox at the same index
        2. The last row in the listbox (if the listbox has a smaller number of rows)
        3. The first row in the following visible listbox
        4. The last row in the preceding visible listbox
        """
        prev_listbox = focused_row.listbox
        prev_index = focused_row.indexx
        destination = None
        if prev_listbox.get_visible():
            at_index = prev_listbox.get_row_at_index(prev_index)
            if at_index:
                destination = at_index
            else:
                destination = prev_listbox.get_last_child()
        if destination:
            destination.grab_focus()
            prev_listbox.select_row(destination)
        else:
            if not self.__focus_next_list_box(prev_listbox):
                self.__focus_previous_list_box(prev_listbox)

    def grab_focus(self) -> None:
        """Pull the focus back to the list."""
        if not self.refocus_selected_row():
            self.move_focus_to_list_top()

    def clear_all_checkboxes(self):
        """Clear all row checkboxes."""
        for listbox in self.__all_listboxes():
            child = listbox.get_first_child()
            while child is not None:
                index_row = child.get_first_child()
                index_row.set_checkbox_value(False)
                child = child.get_next_sibling()

    def update_category_labels(self, index_category: Category) -> None:
        """Update category labels for each row.

        :param Category index_category: The current category
        """
        style = self.__config_manager.index_category_style
        for listbox in self.__visible_listboxes():
            category = None if listbox == self._search_listbox else index_category
            child = listbox.get_first_child()
            while child is not None:
                index_row = child.get_first_child()
                index_row.update_category_label(category, style)
                child = child.get_next_sibling()

    def select_and_focus_note(self, note: Note) -> None:
        """Attempt to select and focus the provided note.

        :param Note note: The note
        """
        self.clear_selections()
        found = False
        for listbox in self.__visible_listboxes():
            child = listbox.get_first_child()
            while child is not None and not found:
                if child.get_first_child().note == note:
                    listbox.select_row(child)
                    child.grab_focus()
                    found = True
                child = child.get_next_sibling()

            if found:
                break

    def focus_next_list_row(self, focused_row: Gtk.Widget) -> None:
        """Focus the next list row, moving to another section if necessary.

        :param Gtk.Widget focused_row: The currently focused row
        """
        listbox = focused_row.get_parent()
        bottom_row = listbox.get_last_child()
        if focused_row == bottom_row:
            self.__focus_next_list_box(listbox)
        else:
            next_row = focused_row.get_next_sibling()
            listbox.select_row(next_row)
            next_row.grab_focus()

    def focus_previous_list_row(self, focused_row: Gtk.Widget) -> None:
        """Focus the previous list row, moving to another section if necessary.

        :param Gtk.Widget focused_row: The currently focused row
        """
        listbox = focused_row.get_parent()
        top_row = listbox.get_first_child()
        if focused_row == top_row:
            self.__focus_previous_list_box(listbox)
        else:
            previous_row = focused_row.get_prev_sibling()
            listbox.select_row(previous_row)
            previous_row.grab_focus()

    def restrict_for_search_by_ids(self, ids: Optional[list[int]], show_all: bool = False) -> None:
        """Restrict by identifiers while searching.

        :param Optional[list[int]] ids: Identifiers
        :param bool show_all: Whether to show the full result set (instead of the most recent with
            an expandable section)
        """
        assert self.__search_model is not None
        if not show_all and ids and len(ids) > self.__search_base_note_count:
            self.__search_full_set = ids
            self._show_more_button.set_visible(True)
            ids = ids[: self.__search_base_note_count]
        else:
            self.__search_full_set = []
            self._show_more_button.set_visible(False)
        self.__search_model.invalidate(ids)

    def show_remaining_search_results(self) -> None:
        """Show remaining search results below first "page"."""
        assert self.__search_model is not None
        if len(self.__search_full_set) > 0:
            self.__search_model.invalidate(self.__search_full_set)
            self.__search_full_set = []
            self._show_more_button.set_visible(False)

    def open_first_search_result(self) -> None:
        """Open the top search result."""
        listbox = self._search_listbox
        if listbox.get_visible():
            row = listbox.get_row_at_index(0)
            if row is None:
                return
            row.grab_focus()
            listbox.select_row(row)
            note = row.get_child().note
            self.emit("note-opened", note)

    def disconnect_older_notes_section(self) -> None:
        """Disconnect listbox and signal for notes older than two months."""
        if not self.older_notes_displayed:
            return
        assert self.__older_notes_model is not None

        self.__older_notes_model.disconnect_by_func(self.__on_older_notes_items_changed)
        self._older_notes_listbox.bind_model(None, self.__create_row, None)
        self.older_notes_displayed = False

    def get_model(self, listbox: Gtk.ListBox) -> Gio.ListModel:
        """Get model for listbox.

        :param Gtk.ListBox listbox: Listbox to fetch for
        :return: Model
        :rtype: Gio.ListModel
        """
        model = None
        for info in self.__main_sections:
            if info.listbox == listbox:
                model = info.model
                break
        if model is not None:
            return model
        elif listbox == self._older_notes_listbox:
            return self.__older_notes_model
        elif listbox == self._search_listbox:
            return self.__search_model

    def iter_visible_notes(self) -> Generator[Note, None, None]:
        """Iterate visible notes.

        :return: Notes
        :rtype: Generator[Note, None, None]
        """
        for listbox in self.__visible_listboxes():
            child = listbox.get_first_child()
            while child is not None:
                row = child.get_first_child()
                row.set_checkbox_value(True)
                yield row.note
                child = child.get_next_sibling()

    @GObject.Property(type=bool, default=False)
    def older_notes_displayed(self) -> bool:
        return self.__older_notes_displayed

    @older_notes_displayed.setter
    def set_older_notes_displayed(self, value: bool) -> None:
        self.__older_notes_displayed = value

    @GObject.Property(type=bool, default=False)
    def searching(self) -> bool:
        return self.__searching

    @searching.setter
    def set_searching(self, value: bool) -> None:
        self.__searching = value

    @GObject.Property(type=bool, default=False)
    def selecting(self) -> bool:
        return self.__selecting

    @selecting.setter
    def set_selecting(self, value: bool) -> None:
        self.__selecting = value

    @GObject.Property(type=bool, default=False)
    def have_more_search_results(self) -> bool:
        return len(self.__search_full_set) > 0

    @GObject.Property(type=bool, default=False)
    def showing_expanded_search_set(self) -> bool:
        """Whether search is active showing a set which has been expanded.

        As in Show More has already been pressed.

        :return: Whether searching and expanded
        :rtype: bool
        """
        return (
            self.searching
            and not self._show_more_button.get_visible()
            and self.__search_model is not None
            and self.__search_model.get_n_items() > self.__search_base_note_count
        )

    def __create_row(self, note: Note, _user_data: Any) -> Gtk.Widget:
        row = IndexRow()
        style = self.__config_manager.index_category_style
        row.populate(note, style)
        row.connect("context-click", self.__on_row_context_click)
        row.connect("checkbox-toggled", self.__on_row_checkbox_toggled)
        self.bind_property("selecting", row, "child-revealed", GObject.BindingFlags.SYNC_CREATE)
        return row

    def __focus_next_list_box(self, previous_listbox: Gtk.ListBox) -> bool:
        found = False
        transferred = False
        for listbox in self.__all_listboxes():
            if not found and listbox == previous_listbox:
                found = True
                continue
            if found and listbox.get_parent().get_visible():
                row = listbox.get_row_at_index(0)
                row.grab_focus()
                listbox.select_row(row)
                previous_listbox.unselect_all()
                transferred = True
                break
        return transferred

    def __focus_previous_list_box(self, previous_listbox: Gtk.ListBox) -> bool:
        found = False
        transferred = False
        for listbox in reversed(list(self.__all_listboxes())):
            if not found and listbox == previous_listbox:
                found = True
                continue
            if found and listbox.get_parent().get_visible():
                row = listbox.get_last_child()
                row.grab_focus()
                listbox.select_row(row)
                previous_listbox.unselect_all()
                transferred = True
                break
        return transferred

    def __on_row_activated(self, listbox: Gtk.ListBox, row: Gtk.ListBoxRow) -> None:
        """Open the note in the selected row in the editor.

        :param Gtk.ListBox listbox: Listbox for activated row
        :param Gtk.ListBoxRow row: The row
        """
        note = row.get_child().note if row else None
        if self.selecting:
            row.get_child().toggle_selected()
        else:
            self.emit("note-opened", note)
            self.__clear_selections_excluding(listbox)

    def __on_row_selected(self, listbox: Gtk.ListBox, row: Optional[Gtk.ListBoxRow]) -> None:
        if row:
            self.__clear_selections_excluding(listbox)

    def __on_row_context_click(self, row: IndexRow) -> None:
        if self.selecting:
            row.toggle_selected()
        else:
            note = row.get_parent().get_first_child().note
            self.select_and_focus_note(note)
            row.set_checkbox_value(True)

    def __on_row_checkbox_toggled(self, row: IndexRow, enabled: bool) -> None:
        if enabled:
            self.emit("note-checkbox-activated", row.note)
        else:
            self.emit("note-checkbox-deactivated", row.note)

    def __on_model_items_changed(
        self,
        model: Gio.ListModel,
        _position: int,
        _removed: int,
        _added: int,
        section: Gtk.Widget,
    ) -> None:
        section.set_visible(not self.searching and len(model) > 0)

    def __on_older_notes_items_changed(
        self, model: Gio.ListModel, _position: int, _removed: int, _added: int
    ) -> None:
        self._older_notes_section.set_visible(
            not self.searching and self.older_notes_displayed and len(model) > 0
        )

    def __on_search_items_changed(
        self, model: Gio.ListModel, _position: int, _removed: int, _added: int
    ) -> None:
        if self.searching and len(model) > 0:
            self.emit("search-model-emptied")

    def __clear_selections_excluding(self, excluding: Optional[Gtk.ListBox]) -> None:
        for listbox in self.__all_listboxes():
            if excluding is None or listbox is not excluding:
                listbox.unselect_all()

    def __add_listbox_keypress_controller(self, listbox: Gtk.ListBox) -> None:
        controller = Gtk.EventControllerKey()
        controller.connect("key-pressed", self.__listbox_keypress_func)
        listbox.add_controller(controller)
        listbox.connect("row-activated", self.__on_row_activated)
        listbox.connect("row-selected", self.__on_row_selected)

    def __bind_older_notes(self) -> None:
        assert self.__older_notes_model is not None
        self._older_notes_listbox.bind_model(self.__older_notes_model, self.__create_row, None)
        self.__older_notes_model.connect("items-changed", self.__on_older_notes_items_changed)

    def __populate_sections_list(self, note_manager: NoteManager) -> None:
        self.__main_sections.append(
            SectionInfo(
                note_manager.favourites_model,
                self._favourites_listbox,
                self._favourites_section,
            )
        )
        self.__main_sections.append(
            SectionInfo(
                note_manager.today_model,
                self._today_listbox,
                self._today_section,
            )
        )
        self.__main_sections.append(
            SectionInfo(
                note_manager.yesterday_model,
                self._yesterday_listbox,
                self._yesterday_section,
            )
        )
        self.__main_sections.append(
            SectionInfo(
                note_manager.week_model,
                self._week_listbox,
                self._week_section,
            )
        )
        self.__main_sections.append(
            SectionInfo(
                note_manager.month_model,
                self._month_listbox,
                self._month_section,
            )
        )
        self.__main_sections.append(
            SectionInfo(
                note_manager.last_month_model,
                self._last_month_listbox,
                self._last_month_section,
            )
        )

    def __visible_listboxes(self) -> Generator[Gtk.ListBox, None, None]:
        for listbox in self.__all_listboxes():
            if listbox.get_parent().get_visible():
                yield listbox

    def __all_listboxes(self) -> Generator[Gtk.ListBox, None, None]:
        for section in self.__main_sections:
            yield section.listbox
        for listbox in self.__extra_listboxes:
            yield listbox

    def __display_older_notes(self) -> bool:
        if self.older_notes_displayed:
            return True
        elif self.__older_notes_model is None:
            return False

        self.older_notes_displayed = True
        self.__bind_older_notes()
        self._older_notes_loader.set_visible(False)

        if len(self.__older_notes_model) > 0:
            self._older_notes_section.set_visible(True)
            return True
        else:
            return False
